import logging
import os
from base64 import b64encode
from datetime import datetime
from getpass import getpass

import requests
from packaging import version
from requests import Session
from requests.adapters import HTTPAdapter, Retry
from requests.cookies import RequestsCookieJar
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from mstrio import config
from mstrio.api import authentication, hooks, misc, projects
from mstrio.helpers import IServerException, VersionException
from mstrio.utils import helper, sessions

logger = logging.getLogger(__name__)


def get_connection(
    workstation_data: dict,
    project_name: str | None = None,
    project_id: str | None = None,
    ssl_verify: bool = False,
) -> 'Connection | None':
    """Connect to environment without providing user's credentials.

    It is possible to provide `project_id` or `project_name` to select
    project. When both `project_id` and `project_name` are `None`,
    project selection is cleared. When both `project_id` and
    `project_name` are provided, `project_name` is ignored.
    Project can be also selected later by calling method
    `select_project` on Connection object.

    Note:
        `ssl_verify` is set to False by default just for the `get_connection`
        function as it is designed for usage inside Workstation.
        When `ssl_verify` is set to False, warning about missing certificate
        verification (InsecureRequestWarning) is disabled.

    Args:
        workstation_data (object): object which is stored in a 'workstationData'
            variable within Workstation
        project_name (str, optional): name of project (aka project)
            to select
        project_id (str, optional): id of project (aka project)
            to select
        ssl_verify (bool, optional): If False (default), does not verify the
            server's SSL certificates

    Returns:
        connection to I-Server or None is case of some error
    """
    if not ssl_verify:
        requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

    try:
        logger.info('Creating connection from Workstation Data object...')
        # get base url from Workstation Data object
        base_url = helper.url_check(workstation_data['defaultEnvironment']['url'])
        # get headers from Workstation Data object
        headers = workstation_data['defaultEnvironment']['headers']
        # get cookies from Workstation Data object
        cookies = workstation_data['defaultEnvironment']['cookies']

        # prepare cookies for request
        jar = RequestsCookieJar()
        for c in cookies:
            cookie_values = {'domain': '', 'path': '/'}
            cookie_values.update(c)
            jar.set(
                cookie_values['name'],
                cookie_values['value'],
                domain=cookie_values['domain'],
                path=cookie_values['path'],
            )
    except Exception as e:
        logger.error(
            f'Some error occurred while preparing data to get identity token: \n{e}'
        )
        return None

    # get identity token
    response = requests.post(
        base_url + '/api/auth/identityToken',
        headers=headers,
        cookies=jar,
        verify=ssl_verify,
    )
    if response.ok:
        # create connection to I-Server
        return Connection(
            base_url,
            identity_token=response.headers['X-MSTR-IdentityToken'],
            project_id=project_id,
            project_name=project_name,
            ssl_verify=ssl_verify,
        )

    logger.error(
        f'HTTP {response.status_code} - {response.reason}, Message {response.text}'
    )


class Connection:
    """Connect to and interact with the MicroStrategy environment.

    Creates a connection object which is used in subsequent requests and
    manages the user's connection with the MicroStrategy REST and Intelligence
    Servers.
    The connection is automatically renewed, or reconnected if server's session
    associated with the connection expires due to inactivity.

    Examples:
        >>> from mstrio import connection
        >>>
        >>> # connect to the environment and chosen project
        >>> conn = connection.Connection(
        >>>     base_url="https://demo.microstrategy.com/MicroStrategyLibrary",
        >>>     username="username",
        >>>     password="password",
        >>>     project_name="MicroStrategy Tutorial"
        >>> )
        >>> # disconnect
        >>> conn.close()

    Attributes:
        base_url: URL of the MicroStrategy REST API server.
        username: Username.
        project_name: Name of the connected MicroStrategy Project.
        project_id: Id of the connected MicroStrategy Project.
        login_mode: Authentication mode. Standard = 1 (default) or LDAP = 16.
        ssl_verify: If True (default), verifies the server's SSL certificates
            with each request.
        user_id: Id of the authenticated user
        user_full_name: Full name of the authenticated user
        user_initials: Initials of the authenticated user
        iserver_version: Version of the I-Server
        web_version: Version of the Web Server
        token: authentication token
        timeout: time after the server's session expires, in seconds
    """

    def __init__(
        self,
        base_url: str,
        username: str | None = None,
        password: str | None = None,
        project_name: str | None = None,
        project_id: str | None = None,
        login_mode: int = 1,
        ssl_verify: bool = True,
        certificate_path: str | None = None,
        proxies: dict | None = None,
        identity_token: str | None = None,
        verbose: bool = True,
    ):
        """Establish a connection with MicroStrategy REST API.

        You can establish connection by either providing set of values
        (`username`, `password`, `login_mode`) or just `identity_token`.

        When both `project_id` and `project_name` are `None`,
        project selection is cleared. When both `project_id`
        and `project_name` are provided, `project_name` is ignored.

        Args:
            base_url (str): URL of the MicroStrategy REST API server.
                Typically of the form:
                "https://<mstr_env>.com/MicroStrategyLibrary/api"
            username (str, optional): Username
            password (str, optional): Password
            project_name (str, optional): Name of the project you intend
                to connect to (case-sensitive). Provide either Project ID
                or Project Name.
            project_id (str, optional): Id of the project you intend to
                connect to (case-sensitive). Provide either Project ID
                or Project Name.
            login_mode (int, optional): Specifies the authentication mode to
                use. Supported authentication modes are: Standard (1)
                (default) or LDAP (16)
            ssl_verify (bool, optional): If True (default), verifies the
                server's SSL certificates with each request
            certificate_path (str, optional): Path to SSL certificate file, if
                None and ssl_verify is True then the certificate will be looked
                for in current working directory
            proxies (dict, optional): Dictionary mapping protocol or protocol
                and host to the URL of the proxy (e.g. {'http': 'foo.bar:3128',
                'http://host.name': 'foo.bar:4012'})
            identity_token (str, optional): Identity token for delegated
                session. Used for connection initialized by GUI.
            verbose (bool, optional): True by default. Controls the amount of
                feedback from the I-Server.
        """

        # set the verbosity globally
        config.verbose = bool(verbose and config.verbose)
        self.base_url = helper.url_check(base_url)
        self.username = username
        self.login_mode = login_mode
        self.certificate_path = certificate_path
        self.identity_token = identity_token
        self._session = self.__configure_session(ssl_verify, certificate_path, proxies)
        self._web_version = None
        self._iserver_version = None
        self._user_id = None
        self._user_full_name = None
        self._user_initials = None
        self.__password = password
        self.last_active = None
        self.timeout = None

        if self.__check_version():
            # save the version of IServer in config file
            config.iserver_version = self.iserver_version
            # delegate identity token or connect and create new session

            if self.identity_token:
                self.delegate()
            else:
                self.connect()

            self.select_project(project_id, project_name)
        else:
            msg = (
                f'This version of mstrio is only supported on MicroStrategy 11.1.0400 '
                f'or higher.\n'
                f'Current Intelligence Server version: {self.iserver_version}\n'
                f'Current MicroStrategy Web version: {self.web_version}'
            )
            logger.warning(msg)
            helper.exception_handler(
                msg='MicroStrategy Version not supported.',
                exception_type=VersionException,
            )

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    def connect(self) -> None:
        """Authenticates the user and creates a new connection with the
        Intelligence Server.

        If an active connection is detected, the session is renewed.
        """
        response = self._renew() if self.token else None

        if response and response.ok:
            self._reset_timeout()
            if config.verbose:
                logger.info(
                    'Connection to MicroStrategy Intelligence Server was renewed.'
                )
        else:
            response = self._login()
            self._reset_timeout()
            self.token = response.headers['X-MSTR-AuthToken']
            self.timeout = self._get_session_timeout()

            if config.verbose:
                logger.info(
                    'Connection to MicroStrategy Intelligence Server has been '
                    'established.'
                )

    renew = connect

    def delegate(self):
        """Delegates identity token to get authentication token and connect to
        MicroStrategy Intelligence Server."""
        response = authentication.delegate(
            self, self.identity_token, whitelist=[('ERR003', 401)]
        )
        if response.ok:
            self._reset_timeout()
            self.token = response.headers['X-MSTR-AuthToken']
            self.timeout = self._get_session_timeout()
            if config.verbose:
                logger.info(
                    'Connection with MicroStrategy Intelligence Server has been '
                    'delegated.'
                )
        else:
            print(
                "Could not share existing connection session, please input credentials:"
            )
            self.__prompt_credentials()
            self.connect()

    def get_identity_token(self) -> str:
        """Create new identity token using existing authentication token."""
        response = authentication.identity_token(self)
        return response.headers['X-MSTR-IdentityToken']

    def validate_identity_token(self) -> bool:
        """Validate the identity token."""
        validate = authentication.validate_identity_token(self, self.identity_token)
        return validate.ok

    def close(self):
        """Closes a connection with MicroStrategy REST API."""
        authentication.logout(connection=self, whitelist=[('ERR009', 401)])

        self._session.close()

        self.token = None

        if config.verbose:
            logger.info(
                'Connection to MicroStrategy Intelligence Server has been closed.'
            )

    def status(self) -> bool:
        """Checks if the session is still alive.

        Raises:
            HTTPError if I-Server behaves unexpectedly
        """
        status = self._status()

        if status.status_code == 200:
            logger.info('Connection to MicroStrategy Intelligence Server is active.')
            return True
        else:
            logger.info(
                'Connection to MicroStrategy Intelligence Server is not active.'
            )
            return False

    def select_project(
        self, project_id: str | None = None, project_name: str | None = None
    ) -> None:
        """Select project for the given connection based on project_id or
        project_name.

        When both `project_id` and `project_name` are `None`, project selection
        is cleared. When both `project_id` and `project_name` are provided,
        `project_name` is ignored.

        Args:
            project_id: id of project to select
            project_name: name of project to select

        Raises:
            ValueError: if project with given id or name does not exist
        """
        if project_id is None and project_name is None:
            self.project_id = None
            self.project_name = None
            if config.verbose:
                logger.info('No project selected.')
            return None

        if project_id and project_name:
            tmp_msg = (
                'Both `project_id` and `project_name` arguments provided. '
                'Selecting project based on `project_id`.'
            )
            helper.exception_handler(msg=tmp_msg, exception_type=Warning)

        _projects = projects.get_projects(connection=self).json()
        if project_id:
            # Find which project name matches the project ID provided
            tmp_projects = helper.filter_list_of_dicts(_projects, id=project_id)
            if not tmp_projects:
                self.project_id, self.project_name = None, None
                tmp_msg = (
                    f"Error connecting to project with id: {project_id}. "
                    "Project with given id does not exist or user has no access."
                )
                raise ValueError(tmp_msg)
        elif project_name:
            # Find which project ID matches the project name provided
            tmp_projects = helper.filter_list_of_dicts(_projects, name=project_name)
            if not tmp_projects:
                self.project_id, self.project_name = None, None
                tmp_msg = (
                    f"Error connecting to project with name: {project_name}. "
                    "Project with given name does not exist or user has no access."
                )
                raise ValueError(tmp_msg)

        self.project_id = tmp_projects[0]['id']
        self.project_name = tmp_projects[0]['name']
        self._session.headers['X-MSTR-ProjectID'] = self.project_id

    @sessions.log_request(logger)
    @sessions.renew_session
    def _request(
        self, method_name: str, url: str, endpoint: str, **kwargs
    ) -> requests.Response:
        if url and endpoint:
            raise ValueError(
                'Either `url` or `endpoint` argument should be provided, not both.'
            )

        if endpoint:
            url = self.base_url + endpoint

        name2method = {
            'GET': self._session.get,
            'POST': self._session.post,
            'PUT': self._session.put,
            'PATCH': self._session.patch,
            'DELETE': self._session.delete,
            'HEAD': self._session.head,
        }
        method = name2method[method_name.upper()]

        return method(url, **kwargs)

    def get(self, url=None, *, endpoint=None, **kwargs):
        """Sends a GET request."""
        return self._request('GET', url, endpoint, **kwargs)

    def post(self, url=None, *, endpoint=None, **kwargs):
        """Sends a POST request."""
        return self._request('POST', url, endpoint, **kwargs)

    def put(self, url=None, *, endpoint=None, **kwargs):
        """Sends a PUT request."""
        return self._request('PUT', url, endpoint, **kwargs)

    def patch(self, url=None, *, endpoint=None, **kwargs):
        """Sends a PATCH request."""
        return self._request('PATCH', url, endpoint, **kwargs)

    def delete(self, url=None, *, endpoint=None, **kwargs):
        """Sends a DELETE request."""
        return self._request('DELETE', url, endpoint, **kwargs)

    def head(self, url=None, *, endpoint=None, **kwargs):
        """Sends a HEAD request."""
        return self._request('HEAD', url, endpoint, **kwargs)

    def _status(self):
        return authentication.session_status(connection=self)

    def _login(self):
        return authentication.login(connection=self)

    def _renew(self):
        return authentication.session_renew(connection=self)

    def _renew_or_reconnect(self):
        response = self._renew() if self.token else None

        if not self.identity_token and (not response or not response.ok):
            response = self._login()
            self.token = response.headers['X-MSTR-AuthToken']

        if response and response.ok:
            self._reset_timeout()

    def _get_session_timeout(self):
        res = self._status()
        return res.json()['timeout'] if res.ok else None

    def _reset_timeout(self):
        self.last_active = datetime.now()

    def _is_session_expired(self):
        return (datetime.now() - self.last_active).seconds > self.timeout

    def _get_authorization(self) -> str:
        self.__prompt_credentials()
        credentials = f"{self.username}:{self.__password}".encode()
        encoded_credential = b64encode(credentials)
        auth = "Basic " + str(encoded_credential, 'utf-8')
        return auth

    def _validate_project_selected(self) -> None:
        if self.project_id is None:
            raise AttributeError(
                "Project not selected. Select project using `select_project` method."
            )

    def __prompt_credentials(self) -> None:
        self.username = self.username or input("Username: ")
        self.__password = self.__password or getpass("Password: ")
        self.login_mode = self.login_mode or input(
            "Login mode (1 - Standard, 16 - LDAP): "
        )

    def __check_version(self) -> bool:
        """Checks version of I-Server and MicroStrategy Web and store these
        variables."""

        def get_server_status():
            json_response = misc.server_status(self).json()
            try:
                iserver_version = json_response["iServerVersion"][:9]
            except KeyError:
                raise IServerException(
                    "I-Server is currently unavailable. Please contact your "
                    "administrator."
                )
            web_version = json_response.get("webVersion")
            web_version = web_version[:9] if web_version else None
            return iserver_version, web_version

        self._iserver_version, self._web_version = get_server_status()
        return version.parse(self.iserver_version) >= version.parse("11.1.0400")

    def __configure_session(
        self,
        verify,
        certificate_path,
        proxies,
        existing_session=None,
        retries=2,
        backoff_factor=0.3,
    ):
        """Creates a shared requests.Session() object with configuration from
        the initialization. Additional parameters change how the HTTPAdapter is
        configured.

        Args:
            existing_session (requests.Session() object, optional): Optional
                existing Session object to configure
            retries (int, optional): number of retries for certain type of
                requests (i.e. get)
            backoff_factor (float, optional):  backoff factor to apply between
                attempts after the second try (most errors are resolved
                immediately by a second try without a delay). urllib3 will sleep
                for: {backoff factor} * (2 ^ ({number of total retries} - 1))
                seconds. If the backoff_factor is 0.1, then sleep() will sleep
                for [0.0s, 0.2s, 0.4s, ...] between retries. It will never be
                longer than Retry. BACKOFF_MAX. By default, backoff is disabled
                (set to 0).
        """
        session = existing_session or Session()
        session.proxies = proxies or {}
        session.verify = self._configure_ssl(verify, certificate_path)

        response_hooks = []
        if config.debug:
            response_hooks.append(hooks.print_url)
        if config.save_responses:
            response_hooks.append(hooks.save_response)
        if response_hooks:
            session.hooks['response'] = response_hooks

        status_forcelist = frozenset([413, 429, 500, 503])  # retry after status

        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)

        return session

    @staticmethod
    def _configure_ssl(ssl_verify, certificate_path):
        def get_certs_from_cwd():
            cert_extensions = ['crt', 'pem', 'p12']
            return [
                file
                for file in os.listdir('.')
                if file.split('.')[-1] in cert_extensions
            ]

        def find_cert_in_cwd(ssl_verify):
            certs = get_certs_from_cwd()
            if certs and ssl_verify:
                return certs[0]
            else:
                return ssl_verify

        if certificate_path:
            return certificate_path
        return find_cert_in_cwd(ssl_verify)

    def __get_user_info(self) -> None:
        response = authentication.get_info_for_authenticated_user(
            connection=self
        ).json()
        self._user_id = response.get("id")
        self._user_full_name = response.get("fullName")
        self._user_initials = response.get("initials")

    @property
    def user_id(self) -> str:
        if not self._user_id:
            self.__get_user_info()
        return self._user_id

    @property
    def user_full_name(self) -> str:
        if not self._user_full_name:
            self.__get_user_info()
        return self._user_full_name

    @property
    def user_initials(self) -> str:
        if not self._user_initials:
            self.__get_user_info()
        return self._user_initials

    @property
    def web_version(self) -> str:
        return self._web_version

    @property
    def iserver_version(self) -> str:
        return self._iserver_version

    @property
    def token(self) -> str:
        return self._session.headers.get('X-MSTR-AuthToken')

    @token.setter
    def token(self, token: str) -> str:
        self._session.headers['X-MSTR-AuthToken'] = token
        return token

    @property
    def headers(self):
        return self._session.headers

    @property
    def environment(self):
        from mstrio.server import Environment

        return Environment(connection=self)



"""This is the demo script to show how administrator can manage documents and
dashboards in users' libraries.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.project_objects import Dashboard, Document, list_documents,list_dashboards
from mstrio.users_and_groups import list_user_groups, list_users, User, UserGroup

from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# list all dashboards and documents within a project to which we have connection
dashboards = list_dashboards(connection=conn)
docs = list_documents(connection=conn)

# Define variables which can be later used in a script
DOCUMENT_NAME = $document_name
DASHBOARD_NAME = $dashboard_name

# get document and dashboard from by name or id and publish them to a library of
# an authenticated user
doc = Document(connection=conn, name=DOCUMENT_NAME)
dashboard = Dashboard(connection=conn, name=DASHBOARD_NAME)
doc.publish()
dashboard.publish()

# Define variables which can be later used in a script
USER_ID_1 = $user_id_1
USER_ID_2 = $user_id_2

# list all users and get 2 of them
users = list_users(connection=conn)
user_1 = User(connection=conn, id=USER_ID_1)
user_2 = User(connection=conn, id=USER_ID_2)

# share one dashboard/document to a given user(s) by passing user object(s)
# or id(s)
dashboard.publish(recipients=user_1)
dashboard.publish(recipients=[USER_ID_1, USER_ID_2])
# analogously we can take away dashboard(s)/document(s) from the library of the
# given user(s)
dashboard.unpublish(recipients=[user_1, user_2])
dashboard.unpublish(recipients=USER_ID_1)

# Define a variable which can be later used in a script
USER_GROUP_NAME = $user_group_name

# list all user groups and get one of them
user_groups_ = list_user_groups(connection=conn)
user_group_ = UserGroup(connection=conn, name=USER_GROUP_NAME)

# add users to given user group
user_group_.add_users(users=[user_1, user_2])

# Define variables which can be later used in a script
DOCUMENT_ID_1 = $document_id_1
DOCUMENT_ID_2 = $document_id_2
DOCUMENT_ID_3 = $document_id_3

# get documents with given ids to give to the user group and users which belong
# to it
docs_to_publish = list_documents(
    connection=conn,
    id=[DOCUMENT_ID_1, DOCUMENT_ID_2, DOCUMENT_ID_3],
)
for doc in docs_to_publish:
    doc.publish(recipients=user_group_)



from mstrio.api import projects as projects_api
from mstrio.connection import Connection


def _prepare_rest_output_for_project_languages(response: dict, path: str) -> dict:
    default_key = 'defaultLanguage' if path == 'data' else 'default'
    default_language_id = response.get(default_key)
    language = response.get('languages').get(default_language_id)
    response[default_key] = {default_language_id: language}
    return response


def get_project_languages(connection: Connection, id: str, path: str):
    """Get languages of the project.

    Args:
        connection (Connection): MicroStrategy REST API connection object
        id (string): Project ID
        path (string): Path to the languages.
            Available values:
                - `data`
                - `metadata`

    Returns:
        List of languages in form of a single dictionary.
    """
    return (
        projects_api.get_project_languages(connection=connection, id=id)
        .json()
        .get(path)
        .get('languages')
    )


def get_default_language(connection: Connection, id: str, path: str):
    """Get default language of the project.

    Args:
        connection (Connection): MicroStrategy REST API connection object
        id (string): Project ID
        path (string): Path to the languages.
            Available values:
                - `data`
                - `metadata`
    """
    return _prepare_rest_output_for_project_languages(
        projects_api.get_project_languages(connection=connection, id=id)
        .json()
        .get(path),
        path,
    ).get('defaultLanguage' if path == 'data' else 'default')


def update_project_languages(connection: Connection, id: str, body: dict, path: str):
    return (
        projects_api.update_project_languages(connection=connection, id=id, body=body)
        .json()
        .get(path)
        .get('languages')
    )


def update_current_mode(connection: Connection, id: str, body: dict):
    return (
        projects_api.update_project_languages(connection=connection, id=id, body=body)
        .json()
        .get('data')
        .get('currentMode')
    )


def update_default_language(connection: Connection, id: str, body: dict):
    return _prepare_rest_output_for_project_languages(
        projects_api.update_project_languages(connection=connection, id=id, body=body)
        .json()
        .get('data'),
        path='data',
    ).get('defaultLanguage')


def get_project_internalization(connection: Connection, id: str):
    response = projects_api.get_project_languages(connection=connection, id=id).json()

    for path in ['data', 'metadata']:
        response[path] = _prepare_rest_output_for_project_languages(
            response=response[path], path=path
        )

    return response



import logging
from dataclasses import dataclass
from enum import auto

from stringcase import snakecase

from mstrio import config
from mstrio.connection import Connection
from mstrio.server.language import list_languages
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable, delete_none_values
from mstrio.utils.response_processors.projects import (
    get_project_languages,
    update_current_mode,
    update_default_language,
    update_project_languages,
)
from mstrio.utils.version_helper import class_version_handler

logger = logging.getLogger(__name__)


@dataclass
class SimpleLanguage(Dictable):
    """Object that specify a Simple Language.

    Attributes:
        id (str): ID of the language.
        name (str, optional): Name of the language.
    """

    id: str
    name: str | None = None

    def __eq__(self, other):
        if not isinstance(other, SimpleLanguage):
            return False

        return self.id == other.id


@dataclass
class DataLocalizationLanguage(SimpleLanguage):
    """Object that specify a Data Localization Language.

    Attributes:
        id (str): ID of the language.
        name (str, optional): Name of the language.
        column (str, optional): Column name for the language.
        table (str, optional): Table name for the language.
        connection_id (str, optional): Connection ID for the language.
    """

    column: str | None = None
    table: str | None = None
    connection_id: str | None = None

    def __eq__(self, other):
        if not isinstance(other, DataLocalizationLanguage):
            return False

        return self.id == other.id

    def __repr__(self):
        return f'DataLocalizationLanguage(id={self.id}, name={self.name})'


class CurrentMode(AutoName):
    """Enum that specify a Current Mode for Data Internationalization."""

    SQL = auto()
    NONE = auto()
    CONNECTION = auto()


def _get_valid_language_ids(
    connection: Connection,
    project_id: str,
    operation: str,
    path: str,
) -> list[str]:
    """Gets valid language IDs for the Project.

    Args:
        connection (Connection): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (str): ID of the project
        operation (str): operation to be performed
            Available values:
                - `add`
                - `remove`
                - `alter`
        path (str): path to target where operation will be performed
            Available values:
                - `data`
                - `metadata`

    Returns:
        A list of valid language IDs for the Project.
    """
    return (
        [language.id for language in list_languages(connection=connection)]
        if operation == 'add'
        else list(get_project_languages(connection, project_id, path))
    )


def _add_language(
    connection: Connection,
    project_id: str,
    languages: list[DataLocalizationLanguage | SimpleLanguage],
    path: str,
) -> list[DataLocalizationLanguage | SimpleLanguage]:
    """Adds languages to the Project.

    Args:
        connection (Connection): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (str): ID of the project
        languages (list[DataLocalizationLanguage | SimpleLanguage]):
            list of languages to be added to the Project
        path (str): path to target where operation will be performed
            Available values:
                - `data`
                - `metadata`

    Returns:
        A list of languages for the Project.
    """
    operation_list = []
    language_type = DataLocalizationLanguage if path == 'data' else SimpleLanguage
    valid_language_ids = _get_valid_language_ids(
        connection=connection, project_id=project_id, operation='add', path=path
    )

    for language in languages:
        if language.id not in valid_language_ids:
            raise ValueError(f"There is no language with the given ID: {language.id}.")

        operation_list.append(
            {
                'op': 'add',
                'path': f'/{path}/languages',
                'value': language_type.to_dict(language),
            }
        )

    body = {'operationList': operation_list}
    body = delete_none_values(body, recursion=True)

    response = update_project_languages(
        connection=connection, id=project_id, body=body, path=path
    )

    return [
        language_type.from_dict({'id': id, **value}, connection)
        for id, value in response.items()
    ]


def _remove_language(
    connection: Connection,
    project_id: str,
    languages: list[DataLocalizationLanguage | SimpleLanguage],
    path: str,
    default_language_id: str,
) -> list[DataLocalizationLanguage | SimpleLanguage]:
    """Removes languages from the Project.

    Args:
        connection (Connection): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (str): ID of the project
        languages (list[DataLocalizationLanguage | SimpleLanguage]):
            list of languages to be added to the Project
        path (str): path to target where operation will be performed
            Available values:
                - `data`
                - `metadata`
    """

    operation_list = []
    language_type = DataLocalizationLanguage if path == 'data' else SimpleLanguage
    valid_language_ids = _get_valid_language_ids(
        connection=connection, project_id=project_id, operation='remove', path=path
    )

    for language in languages:
        if language.id not in valid_language_ids:
            raise ValueError(f"There is no language with the given ID: {language.id}.")
        if language.id == default_language_id:
            raise ValueError(f"Cannot remove default language: {language.id}.")

        operation_list.append(
            {'op': 'remove', 'path': f'/{path}/languages/{language.id}'}
        )

    response = update_project_languages(
        connection=connection,
        id=project_id,
        body={'operationList': operation_list},
        path=path,
    )

    return [
        language_type.from_dict({'id': id, **value}, connection)
        for id, value in response.items()
    ]


@dataclass
@class_version_handler(version='11.3.1200')
class DataLanguageSettings(Dictable):
    """Object that specify a Data Language Settings for a Project.

    Attributes:
        default_language (DataLocalizationLanguage):
            Default language for the Project.
        current_mode (CurrentMode):
            Current mode for data internationalization.
        languages (list[DataLocalizationLanguage]):
            List of languages for the Project.
    """

    _FROM_DICT_MAP = {
        'current_mode': CurrentMode,
        'default_language': lambda source, connection: [
            DataLocalizationLanguage.from_dict({'id': id, **value}, connection)
            for id, value in source.items()
        ][0],
        'languages': lambda source, connection: [
            DataLocalizationLanguage.from_dict({'id': id, **value}, connection)
            for id, value in source.items()
        ],
    }

    default_language: DataLocalizationLanguage
    current_mode: CurrentMode
    languages: list[DataLocalizationLanguage]

    def add_language(self, languages: list[DataLocalizationLanguage]) -> None:
        """Adds languages to the Project.

        Args:
            languages (list[DataLocalizationLanguage]):
                list of languages to be added to the Project
        """

        response = _add_language(
            connection=self._connection,
            project_id=self._project_id,
            languages=languages,
            path='data',
        )
        self.languages = response
        if config.verbose:
            logger.info(
                f"Successfully added languages to the project with ID: "
                f"{self._project_id}"
            )

    def alter_language(
        self, languages: list[DataLocalizationLanguage], path: str
    ) -> None:
        """Alters languages for the Project.

        Args:
            languages (list[DataLocalizationLanguage]):
                list of languages to be altered for the Project
            path (str): path to target where operation will be performed
            Available values:
                - `column`
                - `table`
                - `connectionId`
        """
        if path not in ('column', 'table', 'connectionId'):
            raise ValueError(
                f"Invalid path: {path}. "
                f"Path must be one of: 'column', 'table', 'connectionId'."
            )
        operation_list = []
        valid_language_ids = _get_valid_language_ids(
            connection=self._connection,
            project_id=self._project_id,
            operation='alter',
            path='data',
        )

        for language in languages:
            if language.id not in valid_language_ids:
                raise ValueError(
                    f"There is no language with the given ID: {language.id}."
                )

            operation_list.append(
                {
                    'op': 'replace',
                    'path': f'/data/languages/{language.id}/{path}',
                    'value': getattr(language, snakecase(path)),
                }
            )

        response = update_project_languages(
            connection=self._connection,
            id=self._project_id,
            body={'operationList': operation_list},
            path='data',
        )
        self.languages = [
            DataLocalizationLanguage.from_dict({'id': id, **value}, self._connection)
            for id, value in response.items()
        ]
        if config.verbose:
            logger.info(
                f"Successfully altered languages from the project with ID: "
                f"{self._project_id}"
            )

    def remove_language(self, languages: list[DataLocalizationLanguage]) -> None:
        """Removes languages from the Project.

        Args:
            languages (list[DataLocalizationLanguage]):
                list of languages to be removed from the Project
        """

        response = _remove_language(
            connection=self._connection,
            project_id=self._project_id,
            languages=languages,
            path='data',
            default_language_id=self.default_language.id,
        )

        self.languages = response
        if config.verbose:
            logger.info(
                f"Successfully removed languages from the project with ID: "
                f"{self._project_id}"
            )

    def alter_current_mode(self, mode: str | CurrentMode) -> None:
        """Alters the current mode for data internationalization

        Args:
            mode (str | CurrentMode): mode for data internationalization
                Available values:
                    - `none` | `CurrentMode.NONE`
                    - `sql` | `CurrentMode.SQL`
                    - `connection` | `CurrentMode.CONNECTION`
        """
        mode = mode.value if isinstance(mode, CurrentMode) else mode

        if mode not in ('none', 'sql', 'connection'):
            raise ValueError(
                f"Invalid mode: {mode}. "
                f"Mode must be one of: 'none', 'sql', 'connection'."
            )

        response = update_current_mode(
            connection=self._connection,
            id=self._project_id,
            body={
                'operationList': [
                    {
                        'op': 'replace',
                        'path': '/data/currentMode',
                        'value': mode,
                    }
                ]
            },
        )

        self.current_mode = CurrentMode(response)
        if config.verbose:
            logger.info(
                f"Successfully altered current mode for the project with ID: "
                f"{self._project_id}"
            )

    def alter_default_language(self, language: str | DataLocalizationLanguage) -> None:
        """Alters the default language used by data internationalization

        Args:
            language (str | DataLocalizationLanguage):
                Language or it's ID to be set as default.
        """
        language_id = (
            language.id if isinstance(language, DataLocalizationLanguage) else language
        )
        valid_langauge_ids = _get_valid_language_ids(
            connection=self._connection,
            project_id=self._project_id,
            operation='alter',
            path='data',
        )

        if language_id not in valid_langauge_ids:
            raise ValueError(f"There is no language with the given ID: {language_id}.")

        response = update_default_language(
            connection=self._connection,
            id=self._project_id,
            body={
                'operationList': [
                    {
                        'op': 'replace',
                        'path': '/data/default',
                        'value': language_id,
                    }
                ]
            },
        )

        self.default_language = [
            DataLocalizationLanguage.from_dict({'id': id, **value}, self._connection)
            for id, value in response.items()
        ][0]
        if config.verbose:
            logger.info(
                f"Successfully altered default language "
                f"for the project with ID: {self._project_id}"
            )

    def to_dict(self, camel_case: bool = True) -> dict:
        obj_dict = super().to_dict(camel_case=camel_case)
        if hasattr(obj_dict, '_project_id'):
            obj_dict.pop('_projectId')
        return obj_dict


@dataclass
@class_version_handler(version='11.3.1200')
class MetadataLanguageSettings(Dictable):
    """Object that specify a Metadata Language Settings for a Project.

    Attributes:
        default (SimpleLanguage): Default language for the Project.
        languages (list[SimpleLanguage]): List of languages for the Project.
    """

    _FROM_DICT_MAP = {
        'default': lambda source, connection: [
            SimpleLanguage.from_dict({'id': id, **value}, connection)
            for id, value in source.items()
        ][0],
        'languages': lambda source, connection: [
            SimpleLanguage.from_dict({'id': id, **value}, connection)
            for id, value in source.items()
        ],
    }

    default: SimpleLanguage
    languages: list[SimpleLanguage]

    def add_language(self, languages: list[SimpleLanguage]) -> None:
        """Adds languages to the Project.

        Args:
            languages (list[SimpleLanguage]):
                list of languages to be added to the Project
        """

        response = _add_language(
            connection=self._connection,
            project_id=self._project_id,
            languages=languages,
            path='metadata',
        )
        self.languages = response
        if config.verbose:
            logger.info(
                f"Successfully added languages to the project with ID: "
                f"{self._project_id}"
            )

    def remove_language(self, languages: list[SimpleLanguage]) -> None:
        """Removes languages from the Project.

        Args:
            languages (list[SimpleLanguage]):
                list of languages to be removed from the Project
        """
        response = _remove_language(
            connection=self._connection,
            project_id=self._project_id,
            languages=languages,
            path='metadata',
            default_language_id=self.default.id,
        )
        self.languages = response
        if config.verbose:
            logger.info(
                f"Successfully removed languages from the project with ID: "
                f"{self._project_id}"
            )

    def to_dict(self, camel_case: bool = True) -> dict:
        obj_dict = super().to_dict(camel_case=camel_case)
        if hasattr(obj_dict, '_project_id'):
            obj_dict.pop('_projectId')
        return obj_dict



from typing import TYPE_CHECKING

from requests import Response

from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.helper import exception_handler, response_handler

if TYPE_CHECKING:
    from concurrent.futures import Future

    from mstrio.connection import Connection
    from mstrio.utils.sessions import FuturesSessionWithRenewal


@ErrorHandler(err_msg="Error getting subscription list.")
def list_subscriptions(
    connection, project_id, fields=None, offset=0, limit=-1, error_msg=None
):
    """Get a list of subscriptions.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.get(
        endpoint='/api/subscriptions',
        params={'offset': offset, 'limit': limit, 'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )


def list_subscriptions_async(
    future_session: 'FuturesSessionWithRenewal',
    project_id,
    fields=None,
    offset=0,
    limit=-1,
):
    """Get a list of subscriptions asynchronously.

    Args:
        future_session: Future Session object to call MicroStrategy REST
            Server asynchronously
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).

    Returns:
        Complete Future object.
    """
    params = {'offset': offset, 'limit': limit, 'fields': fields}
    endpoint = '/api/subscriptions'
    headers = {'X-MSTR-ProjectID': project_id}

    return future_session.get(endpoint=endpoint, headers=headers, params=params)


@ErrorHandler(err_msg="Error getting Dynamic Recipient List list.")
def list_dynamic_recipient_lists(
    connection: 'Connection',
    project_id: str,
    offset: int = 0,
    limit: int = -1,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a list of Dynamic Recipient Lists.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.get(
        endpoint='/api/dynamicRecipientLists',
        params={'offset': offset, 'limit': limit, 'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )


def list_dynamic_recipient_lists_async(
    future_session: 'FuturesSessionWithRenewal',
    project_id: str,
    offset: int = 0,
    limit: int = -1,
    fields: str | None = None,
) -> 'Future':
    """Get a list of Dynamic Recipient Lists asynchronously.

    Args:
        future_session: Future Session object to call MicroStrategy REST
            Server asynchronously
        project_id (str): ID of the project
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        Complete Future object.
    """
    return future_session.get(
        endpoint='/api/dynamicRecipientLists',
        headers={'X-MSTR-ProjectID': project_id},
        params={'offset': offset, 'limit': limit, 'fields': fields},
    )


@ErrorHandler(err_msg="Error getting subscription {subscription_id} information.")
def get_subscription(
    connection, subscription_id, project_id, fields=None, error_msg=None
):
    """Get information of a specific subscription for a given project.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        subscription_id (str): ID of the subscription
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.get(
        endpoint=f'/api/subscriptions/{subscription_id}',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error getting Dynamic Recipient List {list_id} information.")
def get_dynamic_recipient_list(
    connection: 'Connection',
    id: str,
    project_id: str,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get information of a specific Dynamic Recipient List for a given project.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): ID of the Dynamic Recipient List
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.get(
        endpoint=f'/api/dynamicRecipientLists/{id}',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error creating new subscription.")
def create_subscription(connection, project_id, body, fields=None, error_msg=None):
    """Create a new subscription.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        body: JSON-formatted body of the new subscription
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.post(
        endpoint='/api/subscriptions',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg="Error creating new Dynamic Recipient List.")
def create_dynamic_recipient_list(
    connection: 'Connection',
    project_id: str,
    body: dict,
    fields: list[str] | None = None,
    error_msg: str | None = None,
) -> Response:
    """Create a new subscription.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        body: JSON-formatted body of the new subscription
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.post(
        endpoint='/api/dynamicRecipientLists',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


def remove_subscription(
    connection, subscription_id, project_id, error_msg=None, exception_type=None
):
    """Remove (Unsubscribe) the subscription using subscription id.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        subscription_id (str): ID of the subscription
        project_id (str): ID of the project
        error_msg (str, optional): Customized error message.
        exception_type (Exception): Instance of Exception or Warning class

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    response = connection.delete(
        endpoint='/api/subscriptions/' + subscription_id,
        headers={'X-MSTR-ProjectID': project_id},
    )
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error unsubscribing Subscription {subscription_id}"
        if exception_type is None:
            response_handler(response, error_msg)
        else:
            exception_handler(error_msg, exception_type)
    return response


@ErrorHandler(err_msg="Error deleting Dynamic Recipient List ID: {id}.")
def remove_dynamic_recipient_list(
    connection: 'Connection', id: str, project_id: str, error_msg: str | None = None
) -> Response:
    """Delete a Dynamic Recipient List.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): ID of the Dynamic Recipient List
        project_id (str): ID of the project
        error_msg (str, optional): Customized error message.
        exception_type (Exception): Instance of Exception or Warning class

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.delete(
        endpoint=f'/api/dynamicRecipientLists/{id}',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error updating subscription {subscription_id}")
def update_subscription(
    connection, subscription_id, project_id, body, fields=None, error_msg=None
):
    """Updates a subscription.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        subscription_id (str): ID of the subscription
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.put(
        endpoint=f'/api/subscriptions/{subscription_id}',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg="Error updating Dynamic Recipient List {list_id}")
def update_dynamic_recipient_list(
    connection: 'Connection',
    id: str,
    project_id: str,
    body: dict,
    fields: list[str] | None = None,
    error_msg: str | None = None,
) -> Response:
    """Updates a Dynamic Recipient List.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): ID of the Dynamic Recipient List
        project_id (str): ID of the project
        body: JSON-formatted body of the new subscription
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.put(
        endpoint=f'/api/dynamicRecipientLists/{id}',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg="Error getting recipients list.")
def available_recipients(
    connection,
    project_id,
    body,
    delivery_type,
    offset=0,
    limit=-1,
    fields=None,
    error_msg=None,
):
    """Get a list of available recipients in shared list, for a given content
    and delivery type, within a given project.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (str): ID of the project
        body (json): {
                    "contents": [
                        {
                        "id": "string",
                        "type": "REPORT"
                        }
                    ]
                    }
        delivery_type (str): Type of the delivery [EMAIL / FILE / PRINTER /
            HISTORY_LIST / CACHE / MOBILE / FTP]
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.


    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.post(
        endpoint='/api/subscriptions/recipients/results',
        params={
            'fields': fields,
            'deliveryType': delivery_type,
            'offset': offset,
            'limit': limit,
        },
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg="Error getting available bursting attributes list.")
def bursting_attributes(
    connection, project_id, content_id, content_type, fields=None, error_msg=None
):
    """Get a list of available attributes for bursting feature, for a given
    content, within a given project. This endpoint returns the name, ID, and
    other information about available attributes.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (str): ID of the project
        content_id (str): ID of the content
        content_type (str): Type of the content [REPORT, DOCUMENT]
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.get(
        endpoint='/api/subscriptions/bursting',
        params={'fields': fields, 'contentId': content_id, 'contentType': content_type},
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error sending subscription {subscription_id}")
def send_subscription(
    connection, subscription_id, project_id, fields=None, error_msg=None
) -> Response:
    """Send the existing subscription immediately.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        subscription_id (str): ID of subscription
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.post(
        endpoint=f'/api/subscriptions/{subscription_id}/send',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )



from mstrio.connection import Connection

from ..schedule import Schedule
from .base_subscription import Subscription
from .content import Content
from .delivery import CacheType, Delivery, LibraryCacheTypes, ShortcutCacheFormat


class CacheUpdateSubscription(Subscription):
    """Class representation of MicroStrategy Cache Update Subscription
    object."""

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        subscription_id: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ):
        """Initialize CacheUpdateSubscription object, populates it with
        I-Server data if id or subscription_id is passed.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.

        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            id (str, optional): ID of the subscription to be initialized, only
                id or subscription_id have to be provided at once, if both are
                provided id will take precedence
            subscription_id (str, optional): ID of the subscription to be
                initialized
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """
        super().__init__(connection, id, subscription_id, project_id, project_name)

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        project_id: str | None = None,
        project_name: str | None = None,
        allow_delivery_changes: bool | None = None,
        allow_personalization_changes: bool | None = None,
        allow_unsubscribe: bool | None = None,
        send_now: bool | None = None,
        owner_id: str | None = None,
        schedules: str | list[str] | Schedule | list[Schedule] | None = None,
        contents: Content | None = None,
        recipients: list[dict] | list[str] | None = None,
        delivery: Delivery | dict | None = None,
        delivery_expiration_date: str | None = None,
        delivery_expiration_timezone: str | None = None,
        contact_security: bool = True,
        cache_cache_type: CacheType | str | None = None,
        cache_shortcut_cache_format: ShortcutCacheFormat | str | None = None,
        cache_library_cache_types: list[LibraryCacheTypes | str] | None = None,
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: str | None = None,
    ) -> "CacheUpdateSubscription":
        """Creates a new cache update subscription.

        Args:
            connection (Connection): a MicroStrategy connection object
            name (str): name of the subscription,
            project_id (str, optional): project ID,
            project_name (str, optional): project name,
            allow_delivery_changes (bool, optional): whether the recipients can
                change the delivery of the subscription,
            allow_personalization_changes (bool, optional): whether the
                recipients can personalize the subscription,
            allow_unsubscribe (bool, optional): whether the recipients can
                unsubscribe from the subscription,
            send_now (bool, optional): indicates whether to execute the
                subscription immediately,
            owner_id (str, optional): ID of the subscription owner, by default
                logged in user ID,
            schedules (str | list[str] | Schedule | list[Schedule], optional):
                Schedules IDs or Schedule objects,
            contents (Content, optional): The content settings.
            recipients (list[str] | list[dict], optional): list of recipients
                IDs or dicts,
            delivery (Delivery | dict, optional): delivery object or dict
            delivery_expiration_date (str, optional): expiration date of the
                subscription, format should be yyyy - MM - dd,
            delivery_expiration_timezone (str, optional): expiration timezone
                of the subscription, example value 'Europe/London'
            contact_security (bool): whether to use contact security for each
                contact group member
            cache_cache_type (CacheType | str, optional):
                [RESERVED, SHORTCUT, SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format (ShortcutCacheFormat | str, optional):
                [RESERVED, JSON, BINARY, BOTH]
            cache_library_cache_types (list[LibraryCacheTypes | str], optional):
                Set of library cache types, available types can be
                web, android, ios
            cache_reuse_dataset_cache (bool): Whether to reuse dataset cache
            cache_is_all_library_users (bool): Whether for all library users
            delivery_notification_enabled (bool): Whether notification is
                enabled, notification applies to cache
            delivery_personal_notification_address_id (str, optional):
                Notification details
        """
        notification_address_id = delivery_personal_notification_address_id

        return super()._Subscription__create(
            connection=connection,
            name=name,
            project_id=project_id,
            project_name=project_name,
            allow_delivery_changes=allow_delivery_changes,
            allow_personalization_changes=allow_personalization_changes,
            allow_unsubscribe=allow_unsubscribe,
            send_now=send_now,
            owner_id=owner_id,
            schedules=schedules,
            contents=contents,
            recipients=recipients,
            delivery=delivery,
            delivery_mode=Delivery.DeliveryMode.CACHE,
            delivery_expiration_date=delivery_expiration_date,
            delivery_expiration_timezone=delivery_expiration_timezone,
            contact_security=contact_security,
            cache_cache_type=cache_cache_type or CacheType.RESERVED,
            cache_shortcut_cache_format=cache_shortcut_cache_format
            or ShortcutCacheFormat.RESERVED,
            cache_library_cache_types=cache_library_cache_types
            or [LibraryCacheTypes.WEB],
            cache_reuse_dataset_cache=cache_reuse_dataset_cache,
            cache_is_all_library_users=cache_is_all_library_users,
            delivery_notification_enabled=delivery_notification_enabled,
            delivery_personal_notification_address_id=notification_address_id,
        )

    def alter(
        self,
        name: str | None = None,
        allow_delivery_changes: bool | None = None,
        allow_personalization_changes: bool | None = None,
        allow_unsubscribe: bool | None = None,
        send_now: bool | None = None,
        owner_id: str | None = None,
        schedules: str | list[str] | Schedule | list[Schedule] | None = None,
        contents: Content | None = None,
        recipients: list[dict] | list[str] | None = None,
        delivery: Delivery | dict | None = None,
        custom_msg: str | None = None,
        delivery_expiration_date: str | None = None,
        delivery_expiration_timezone: str | None = None,
        contact_security: bool = True,
        cache_cache_type: CacheType | str | None = None,
        cache_shortcut_cache_format: ShortcutCacheFormat | str | None = None,
        cache_library_cache_types: list[LibraryCacheTypes | str] | None = None,
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: str | None = None,
    ):
        """Alter the subscription.

        Args:
            name (str): name of the subscription,
            allow_delivery_changes (bool, optional): whether the recipients can
                change the delivery of the subscription,
            allow_personalization_changes (bool, optional): whether the
                recipients can personalize the subscription,
            allow_unsubscribe (bool, optional): whether the recipients can
                unsubscribe from the subscription,
            send_now (bool, optional): indicates whether to execute the
                subscription immediately,
            owner_id (str, optional): ID of the subscription owner, by default
                logged in user ID,
            schedules (str | list[str] | Schedule | list[Schedule], optional):
                Schedules IDs or Schedule objects,
            contents (Content, optional): The content settings.
            recipients (list[str] | list[dict], optional): list of recipients
                IDs or dicts,
            delivery (Delivery | dict, optional): delivery object or dict
            delivery_expiration_date (str, optional): expiration date of the
                subscription, format should be yyyy - MM - dd,
            delivery_expiration_timezone (str, optional): expiration timezone
                of the subscription
            contact_security (bool): whether to use contact security for each
                contact group member
            cache_cache_type (CacheType | str, optional):
                [RESERVED, SHORTCUT, SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format (ShortcutCacheFormat | str, optional):
                [RESERVED, JSON, BINARY, BOTH]
            cache_library_cache_types (list[LibraryCacheTypes | str], optional):
                Set of library cache types, available types can be
                web, android, ios
            cache_reuse_dataset_cache (bool): Whether to reuse dataset cache
            cache_is_all_library_users (bool): Whether for all library users
            delivery_notification_enabled (bool): Whether notification is
                enabled, notification applies to cache
            delivery_personal_notification_address_id (str, optional):
                Notification details
        """
        notification_address_id = delivery_personal_notification_address_id

        return super().alter(
            name=name,
            allow_delivery_changes=allow_delivery_changes,
            allow_personalization_changes=allow_personalization_changes,
            allow_unsubscribe=allow_unsubscribe,
            send_now=send_now,
            owner_id=owner_id,
            schedules=schedules,
            contents=contents,
            recipients=recipients,
            delivery=delivery,
            custom_msg=custom_msg,
            delivery_expiration_date=delivery_expiration_date,
            delivery_expiration_timezone=delivery_expiration_timezone,
            contact_security=contact_security,
            cache_cache_type=cache_cache_type or CacheType.RESERVED,
            cache_shortcut_cache_format=cache_shortcut_cache_format
            or ShortcutCacheFormat.RESERVED,
            cache_library_cache_types=cache_library_cache_types
            or [LibraryCacheTypes.WEB],
            cache_reuse_dataset_cache=cache_reuse_dataset_cache,
            cache_is_all_library_users=cache_is_all_library_users,
            delivery_notification_enabled=delivery_notification_enabled,
            delivery_personal_notification_address_id=notification_address_id,
        )



import logging
from enum import Enum
from operator import itemgetter
from typing import TYPE_CHECKING

import requests
from pandas.core.frame import DataFrame
from tqdm.auto import tqdm

from mstrio import config
from mstrio.api import cubes, datasets
from mstrio.connection import Connection
from mstrio.helpers import NotSupportedError
from mstrio.object_management.search_operations import SearchPattern, full_search
from mstrio.project_objects.datasets import cube_cache
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.entity import DeleteMixin, Entity, VldbMixin
from mstrio.utils.filter import Filter
from mstrio.utils.helper import (
    choose_cube,
    exception_handler,
    fallback_on_timeout,
    get_parallel_number,
    get_valid_project_id,
    response_handler,
    sort_object_properties,
)
from mstrio.utils.parser import Parser
from mstrio.utils.response_processors import cubes as cube_processors
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.sessions import FuturesSessionWithRenewal
from mstrio.utils.translation_mixin import TranslationMixin

if TYPE_CHECKING:
    from .cube_cache import CubeCache
    from .olap_cube import OlapCube
    from .super_cube import SuperCube

logger = logging.getLogger(__name__)


class CubeStates(Enum):
    ACTIVE = 2
    DIRTY = 16
    DIRTY_INFO = 8
    FOREIGN = 2048
    IMPORTED = 1024
    LOAD_PENDING = 128
    LOADED = 32
    PENDING_FOR_ENGINE = 512
    PERSISTED = 4
    PROCESSING = 1
    READY = 64
    RESERVED = 0
    UNLOAD_PENDING = 256
    UNKNOWN1 = 4096
    UNKNOWN2 = 8192
    UNKNOWN3 = 16384

    @classmethod
    def show_status(cls, status: int) -> list[str]:
        """Show states of a cube calculated from numerical value of its state.
        Additionally list of those states' names is returned.

        Args:
            status (integer): numerical value of cube's status

        Returns:
            List with names of cube's states.
        """

        def parse_cube_status_bin_to_list(status_bin: int) -> list[str]:
            output = []
            # sort states from enum based on its value and return as a table of
            # tuples (value, name)
            states = sorted(((s.value, s.name) for s in cls), key=lambda x: -x[0])
            for state in states:
                if status_bin >= state[0]:
                    status_bin -= state[0]
                    output.append(state[1])
            return output

        states = parse_cube_status_bin_to_list(status)
        output = []
        for state in states:
            if 'UNKNOWN' in state:
                continue
            output.append(' '.join(s.capitalize() for s in state.split('_')))
        print(', '.join(output))
        return output


def list_all_cubes(
    connection: Connection,
    name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> 'list[OlapCube | SuperCube] | list[dict]':
    """Get list of Cube objects (OlapCube or SuperCube) or dicts with them.
    Optionally filter cubes by specifying 'name'.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name = ?onny will return Sonny and Tonny

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of cubes being searched
        search_pattern (SearchPattern enum or int, optional): pattern to search
            for, such as Begin With or Contains. Possible values are available
            in ENUM mstrio.object_management.SearchPattern.
            Default value is BEGIN WITH (4).
        project_id (string, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns SuperCube/OlapCube objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        **filters: Available filter parameters: ['id', 'name', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'owner',
            'ext_type', 'view_media', 'certified_info']

    Returns:
        list with OlapCubes and SuperCubes or list of dictionaries
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    objects_ = full_search(
        connection,
        project=project_id,
        name=name,
        object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE],
        pattern=search_pattern,
        limit=limit,
        **filters,
    )
    if to_dictionary:
        return objects_
    else:
        all_cubes = []
        for object_ in objects_:
            cube_subtype = object_['subtype']
            if cube_subtype == int(ObjectSubTypes.OLAP_CUBE):
                from .olap_cube import OlapCube

                all_cubes.append(
                    OlapCube.from_dict(
                        source=object_, connection=connection, with_missing_value=True
                    )
                )
            elif cube_subtype == int(ObjectSubTypes.SUPER_CUBE):
                from .super_cube import SuperCube

                all_cubes.append(
                    SuperCube.from_dict(
                        source=object_, connection=connection, with_missing_value=True
                    )
                )
        return all_cubes


def load_cube(
    connection: Connection,
    cube_id: str | None = None,
    cube_name: str | None = None,
    folder_id: str | None = None,
    instance_id: str | None = None,
) -> 'OlapCube | SuperCube | list[OlapCube | SuperCube]':
    """Load single cube specified by either 'cube_id' or both 'cube_name' and
    'folder_id'.

    It is also possible to load cube by providing only `cube_name`, but in that
    case we may retrieve more than one cube as cube's name is unique only within
    a folder.

    `instance_id` is used only when a single cube is retrieved.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        cube_id(string, optional): ID of cube
        cube_name(string, optional): name of cube
        folder_id(string, optional): ID of folder in which the cube is stored
        instance_id (str, optional): Identifier of an instance if cube instance
            has been already initialized. Will be used if only one cube is
            found. `None` by default.

    Returns:
         object of type OlapCube or SuperCube based on the subtype of a cube
         returned from metadata.

    Raises:
        ValueError when neither `cube_id` nor `cube_name` are provided.
    """
    connection._validate_project_selected()

    if cube_id:
        if cube_name:
            msg = (
                "Both `cube_id` and `cube_name` provided. Loading cube based on "
                "`cube_id`."
            )
            exception_handler(msg, Warning, False)
        elif folder_id:
            msg = (
                "Both `cube_id` and `folder_id` provided. "
                "Loading cube based on `cube_id` from all folders."
            )
            exception_handler(msg, Warning, False)
        objects_ = full_search(
            connection,
            project=connection.project_id,
            object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE],
            pattern=SearchPattern.EXACTLY,
            id=cube_id,
        )
    elif not cube_name:
        msg = "Specify either `cube_id` or `cube_name`."
        raise ValueError(msg)
    else:  # getting cube by `cube_name` and optionally `folder_id`
        objects_ = full_search(
            connection,
            project=connection.project_id,
            name=cube_name,
            object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE],
            pattern=SearchPattern.EXACTLY,
            root=folder_id,
        )

    ret_cubes = []
    for object_ in objects_:
        object_ = (
            object_ if len(objects_) > 1 else {**object_, 'instance_id': instance_id}
        )

        ret_cubes.append(choose_cube(connection, object_))
        ret_cubes = [tmp for tmp in ret_cubes if tmp]  # remove `None` values

    if len(ret_cubes) == 0:
        exception_handler("Cube was not found.", Warning, False)
        return None
    elif len(ret_cubes) == 1:
        return ret_cubes[0]
    else:
        exception_handler(
            f"More than one cube with name {cube_name} was loaded.", Warning, False
        )
        return ret_cubes


class _Cube(Entity, VldbMixin, DeleteMixin, TranslationMixin):
    """Access, filter, publish, and extract data from MicroStrategy in-memory
    cubes.

    Create a Cube object to load basic information on a cube dataset. Specify
    subset of cube to be fetched through `Cube.apply_filters()` and
    `Cube.clear_filters()`. Fetch dataset through `Cube.to_dataframe()` method.

    Attributes:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        id: Identifier of a pre-existing cube containing the required data.
        instance_id (str): Identifier of an instance if cube instance has been
            already initialized, NULL by default.
    """

    _OBJECT_TYPE = ObjectTypes.REPORT_DEFINITION
    _OBJECT_SUBTYPE = None
    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'hidden',
            'source',
            'comments',
            'template_info',
            'target_info',
            'hidden',
            'comments',
        ): objects_processors.get_info,
        ('server_mode', 'size', 'path', 'status', 'owner_id'): cube_processors.get_info,
    }
    _API_PATCH: dict = {
        'comments': (objects_processors.update, 'partial_put'),
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'certified_info': CertifiedInfo.from_dict,
    }
    _SIZE_LIMIT = 10000000  # this sets desired chunk size in bytes

    def __init__(
        self,
        connection: Connection,
        id: str,
        name: str | None = None,
        instance_id: str | None = None,
        parallel: bool = True,
        progress_bar: bool = True,
    ):
        """Initialize an instance of a cube by its id.

        Note:
            Parameter `name` is not used when fetching. `id` is always used to
            uniquely identify cube.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of a pre-existing cube containing
                the required data.
            name (str): Name of a cube.
            instance_id (str): Identifier of an instance if cube instance has
                been already initialized, None by default.
            parallel (bool, optional): If True (default), utilize optimal number
                of threads to increase the download speed. If False, this
                feature will be disabled.
            progress_bar(bool, optional): If True (default), show the download
                progress bar.
        """
        super().__init__(
            connection,
            id,
            name=name,
            instance_id=instance_id,
            parallel=parallel,
            progress_bar=progress_bar,
            subtype=self._OBJECT_SUBTYPE,
        )
        connection._validate_project_selected()
        self._get_definition()

    def _init_variables(self, default_value, **kwargs):
        super()._init_variables(default_value=default_value, **kwargs)
        self.instance_id = kwargs.get('instance_id')
        self._parallel = kwargs.get('parallel', True)
        self._initial_limit = 1000
        self._progress_bar = bool(
            kwargs.get("progress_bar", True) and config.progress_bar
        )
        self._table_definition = {}
        self._dataframe = None
        self._dataframes = []
        self._attr_elements = None
        # these properties were not fetched from self.__definition() and will be
        # lazily fetched when calling properties `attributes` or `metrics`
        self._attributes = []
        self._metrics = []
        self._row_counts = []
        self._table_names = []
        self.__definition_retrieved = False
        # these properties were not fetched from self.__info() and all will be
        # lazily fetched when calling any of properties: `owner_id`, `path`,
        # `size`, `status`
        self.owner_id = default_value
        self.path = default_value
        self.server_mode = default_value
        self.size = default_value
        self.status = default_value
        self.__filter = None

        # caches will be lazily retrieved from I-Server when calling  for cube's
        # property `caches`
        self._caches = None

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        abbreviation: str | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
    ):
        """Alter Cube properties.

        Args:
            name: new name of the Dataset
            description: new description of the Dataset
            abbreviation: new abbreviation of the Dataset
            hidden: Specifies whether the metric is hidden
            comments: new long description of the Dataset
        """
        func = self.alter
        args = func.__code__.co_varnames[: func.__code__.co_argcount]
        defaults = func.__defaults__  # type: ignore
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def get_caches(self) -> list['CubeCache']:
        """Get list of caches of the cube.

        Returns:
            list of caches of the cube
        """
        self._caches = cube_cache.list_cube_caches(self._connection, cube_id=self._id)
        return self._caches

    def to_dataframe(self, limit: int | None = None, multi_df: bool = False):
        """Extract contents of a cube into a Pandas `DataFrame`.

        Args:
            limit (None or int, optional): Used to control data extract
                behavior. By default (None) the limit is calculated
                automatically, based on an optimized physical size of one chunk.
                Setting limit manually will force the number of rows per chunk.
                Depending on system resources, a higher limit (e.g. 50,000) may
                reduce the total time required to extract the entire dataset.
            multi_df (bool, optional): If True, return a list of data frames
                resembling the table structure of the cube. If False (default),
                returns one data frame.

        Returns:
            Pandas Data Frame containing the cube contents
        """
        if limit:
            self._initial_limit = limit

        if self.instance_id is None:
            res = self.__create_cube_instance(self._initial_limit)
        else:
            # try to get first chunk from already initialized instance of cube,
            # if not possible, initialize new instance
            try:
                res = cubes.cube_instance_id(
                    connection=self.connection,
                    cube_id=self._id,
                    instance_id=self.instance_id,
                    offset=0,
                    limit=self._initial_limit,
                )
            except requests.HTTPError:
                res = self.__create_cube_instance(self._initial_limit)

        # Gets the pagination totals and instance_id from the response object
        _instance = res.json()
        self.instance_id = _instance['instanceId']
        paging = _instance['data']['paging']

        # initialize parser and process first response
        p = Parser(response=_instance, parse_cube=True)
        p.parse(response=_instance)

        # If there are more rows to fetch, fetch them
        if paging['current'] != paging['total']:
            if not limit:
                limit = max(
                    1000,
                    int((self._initial_limit * self._SIZE_LIMIT) / len(res.content)),
                )
            # Count the number of additional iterations
            it_total = int((paging['total'] - self._initial_limit) / limit) + (
                (paging['total'] - self._initial_limit) % limit != 0
            )

            if self._parallel and it_total > 1:
                threads = get_parallel_number(it_total)
                with FuturesSessionWithRenewal(
                    connection=self._connection, max_workers=threads
                ) as session:
                    fetch_pbar = tqdm(
                        desc="Downloading",
                        total=it_total + 1,
                        disable=(not self._progress_bar),
                    )
                    future = self.__fetch_chunks_future(
                        session, paging, self.instance_id, limit
                    )
                    fetch_pbar.update()
                    for i, f in enumerate(future, start=1):
                        response = f.result()
                        if not response.ok:
                            response_handler(response, "Error getting cube contents.")
                        fetch_pbar.update()
                        fetch_pbar.set_postfix(
                            rows=str(
                                min(self._initial_limit + i * limit, paging['total'])
                            )
                        )
                        p.parse(response.json())
                    fetch_pbar.close()
            else:
                self.__fetch_chunks(p, paging, it_total, self.instance_id, limit)

        # return parsed data as a data frame
        self._dataframe = p.dataframe
        # split dataframe to dataframes matching tables in Cube
        if multi_df:
            if p.has_multiform_attributes():
                raise NotSupportedError(
                    "Splitting into multiple dataframes is not supported for cubes "
                    "containing multiform attributes."
                )
            # split dataframe to dataframes matching tables in Cube
            self._dataframes = [
                self._dataframe[columns].copy()
                for _, columns in self.__multitable_definition().items()
            ]
            return self._dataframes
        else:
            return self._dataframe

    def __fetch_chunks_future(self, future_session, pagination, instance_id, limit):
        """Fetch add'l rows from this object instance from the Intelligence
        Server."""
        return [
            cubes.cube_instance_id_coroutine(
                future_session,
                connection=self._connection,
                cube_id=self._id,
                instance_id=instance_id,
                offset=_offset,
                limit=limit,
            )
            for _offset in range(self._initial_limit, pagination['total'], limit)
        ]

    def __fetch_chunks(self, parser, pagination, it_total, instance_id, limit):
        """Fetch add'l rows from this object instance from the Intelligence
        Server."""
        with tqdm(
            desc="Downloading", total=it_total + 1, disable=(not self._progress_bar)
        ) as fetch_pbar:
            fetch_pbar.update()
            for _offset in range(self._initial_limit, pagination['total'], limit):
                response = cubes.cube_instance_id(
                    connection=self.connection,
                    cube_id=self._id,
                    instance_id=instance_id,
                    offset=_offset,
                    limit=limit,
                )
                fetch_pbar.update()
                fetch_pbar.set_postfix(
                    rows=str(min(_offset + limit, pagination['total']))
                )
                parser.parse(response=response.json())

    def __create_cube_instance(self, limit):
        inst_pbar = tqdm(
            desc="Initializing an instance of a cube. Please wait...",
            bar_format='{desc}',
            leave=False,
            ncols=280,
            disable=(not self._progress_bar),
        )
        # Request a new instance, set instance id
        response = cubes.cube_instance(
            connection=self._connection,
            cube_id=self._id,
            body=self._prepare_instance_body(),
            offset=0,
            limit=limit or self._initial_limit,
        )
        inst_pbar.close()
        return response

    def apply_filters(
        self,
        attributes: list | None = None,
        metrics: list | None = None,
        attr_elements: list | None = None,
        operator: str = 'In',
    ) -> None:
        """Apply filters on the cube's objects.

        Filter by attributes, metrics and attribute elements.

        Args:
            attributes (list or None, optional): ids of attributes to be
                included in the filter. If list is empty, no attributes will be
                selected and metric data will be aggregated.
            metrics (list or None, optional): ids of metrics to be included in
                the filter. If list is empty, no metrics will be selected.
            attr_elements (list or None, optional): attribute elements to be
                included in the filter.
            operator (str, optional): a str flag used to specify if the
                attribute elements selected inside the filter should be included
                or excluded. Allowed values are: 'In', 'NotIn'.
        """
        params = [attributes, metrics, attr_elements]
        filtering_is_requested = bool(not all(el is None for el in params))

        if filtering_is_requested:
            self._instance_config._clear(
                attributes=attributes, metrics=metrics, attr_elements=attr_elements
            )
            self._instance_config.operator = operator
            self._select_attribute_filter_conditionally(attributes)
            self._select_metric_filter_conditionally(metrics)
            self._select_attr_el_filter_conditionally(attr_elements)
            # Clear instance, to generate new with new filters
            self.instance_id = None

    def _select_attribute_filter_conditionally(self, attributes_filtered):
        if attributes_filtered:
            self._instance_config._select(object_id=attributes_filtered)
        elif attributes_filtered is not None:
            self._instance_config.attr_selected = []

    def _select_metric_filter_conditionally(self, metrics_filtered):
        if metrics_filtered:
            self._instance_config._select(object_id=metrics_filtered)
        elif metrics_filtered is not None:
            self._instance_config.metr_selected = []

    def _select_attr_el_filter_conditionally(self, attr_el_filtered):
        if attr_el_filtered is not None:
            self._instance_config._select_attr_el(element_id=attr_el_filtered)

    def clear_filters(self) -> None:
        """Clear previously set filters, allowing all attributes, metrics, and
        attribute elements to be retrieved."""

        self._instance_config._clear()

        # once again remove Row Count metrics
        metrics_ids = [metric_id['id'] for metric_id in self.metrics]
        self._instance_config._select(metrics_ids)
        # Clear instance, to generate new with new filters
        self.instance_id = None

    def refresh_status(self) -> None:
        """Refresh cube's status and show which states it represents."""
        res = cubes.status(self._connection, self._id)
        if res.ok:
            self.status = int(res.headers['X-MSTR-CubeStatus'])

    def show_status(self) -> list[str]:
        """Show which states are represented by cube's status."""
        return CubeStates.show_status(self.status)

    def list_properties(self) -> dict:
        """List all properties of the object."""

        attributes = {
            key: self.__dict__[key] for key in self.__dict__ if not key.startswith('_')
        }
        attributes = {
            **attributes,
            'id': self.id,
            'instance_id': self.instance_id,
            'type': self.type,
            'subtype': self.subtype,
            'ext_type': self.ext_type,
            'date_created': self.date_created,
            'date_modified': self.date_modified,
            'version': self.version,
            'owner': self.owner,
            'view_media': self.view_media,
            'ancestors': self.ancestors,
            'certified_info': self.certified_info,
            'acg': self.acg,
            'acl': self.acl,
            'size': self.size,
            'status': self.status,
            'path': self.path,
            'attributes': self.attributes,
            'metrics': self.metrics,
        }
        return {
            key: attributes[key]
            for key in sorted(attributes, key=sort_object_properties)
        }

    def _get_definition(self) -> None:
        """Get the definition of a cube, including attributes and metrics.

        Implements GET /v2/cubes/<cube_id>.
        """
        if self._id is not None:
            _definition = cubes.cube_definition(
                connection=self._connection, id=self._id
            ).json()
            full_attributes = _definition['definition']['availableObjects'][
                'attributes'
            ]
            full_metrics = _definition['definition']['availableObjects']['metrics']
            self._attributes = [
                {'name': attr['name'], 'id': attr['id']} for attr in full_attributes
            ]
            self._metrics = [
                {'name': metr['name'], 'id': metr['id']} for metr in full_metrics
            ]

            self._table_names = self.__multitable_definition().keys()
            row_counts = [
                f'Row Count - {table_name}' for table_name in self._table_names
            ]
            self._row_counts = list(
                filter(lambda x: x['name'] in row_counts, self._metrics)
            )
            self.__remove_row_count()
            # for lazy fetch properties
            self.__definition_retrieved = True

    def __multitable_definition(self):
        """Return all tables names and columns as a dictionary."""
        if not self._table_definition:
            res_tables = datasets.dataset_definition(
                connection=self._connection,
                id=self._id,
                fields=['tables', 'columns'],
                whitelist=[('ERR001', 500)],
            )
            if res_tables.ok:
                ds_definition = res_tables.json()
                for table in ds_definition['result']['definition']['availableObjects'][
                    'tables'
                ]:
                    column_list = [
                        column['columnName']
                        for column in ds_definition['result']['definition'][
                            'availableObjects'
                        ]['columns']
                        if table['name'] == column['tableName']
                    ]
                    self._table_definition[table['name']] = column_list
        return self._table_definition

    def __remove_row_count(self):
        """Remove all Row Count metrics from cube."""
        row_counts = list(map(itemgetter('name'), self._row_counts))
        self._metrics = list(
            filter(lambda x: x['name'] not in row_counts, self._metrics)
        )

    def __get_attr_elements(self, limit=50000):
        """Get elements of report attributes synchronously.

        Implements GET /reports/<report_id>/attributes/<attribute_id>/elements.
        """

        def fetch_for_attribute(attribute):
            @fallback_on_timeout()
            def fetch_for_attribute_given_limit(limit):
                response = cubes.cube_single_attribute_elements(
                    connection=self._connection,
                    cube_id=self._id,
                    attribute_id=attribute['id'],
                    offset=0,
                    limit=limit,
                )
                # Get total number of rows from headers.
                total = int(response.headers['x-mstr-total-count'])
                # Get attribute elements from the response.
                elements = response.json()

                # If total number of elements is bigger than the chunk size
                # (limit), fetch them incrementally.
                for _offset in range(limit, total, limit):
                    response = cubes.cube_single_attribute_elements(
                        connection=self._connection,
                        cube_id=self._id,
                        attribute_id=attribute['id'],
                        offset=_offset,
                        limit=limit,
                    )
                    elements.extend(response.json())

                # Return attribute data.
                return {
                    'attribute_name': attribute['name'],
                    'attribute_id': attribute['id'],
                    'elements': elements,
                }

            return fetch_for_attribute_given_limit(limit)[0]

        attr_elements = []
        if self.attributes:
            pbar = tqdm(
                self.attributes,
                desc="Loading attribute elements",
                leave=False,
                disable=(not self._progress_bar),
            )
            attr_elements = [fetch_for_attribute(attribute) for attribute in pbar]
            pbar.close()

        return attr_elements

    def __get_attr_elements_async(self, limit=50000):
        """Get attribute elements.

        Implements GET /cubes/<cube_id>/attributes/<attribute_id>/elements.
        """

        attr_elements = []
        if self.attributes:
            threads = get_parallel_number(len(self.attributes))
            with FuturesSessionWithRenewal(
                connection=self._connection, max_workers=threads
            ) as session:
                # Fetch first chunk of attribute elements.
                futures = self.__fetch_attribute_elements_chunks(session, limit)
                pbar = tqdm(
                    futures,
                    desc="Loading attribute elements",
                    leave=False,
                    disable=(not self._progress_bar),
                )
                for i, future in enumerate(pbar):
                    attr = self.attributes[i]
                    response = future.result()
                    if not response.ok:
                        response_handler(
                            response, f"Error getting attribute {attr['name']} elements"
                        )
                    elements = response.json()
                    # Get total number of rows from headers.
                    total = int(response.headers['x-mstr-total-count'])
                    for _offset in range(limit, total, limit):
                        response = cubes.cube_single_attribute_elements(
                            connection=self._connection,
                            cube_id=self._id,
                            attribute_id=attr['id'],
                            offset=_offset,
                            limit=limit,
                        )
                        elements.extend(response.json())
                    # Append attribute data to the list of attributes.
                    attr_elements.append(
                        {
                            'attribute_name': attr['name'],
                            'attribute_id': attr['id'],
                            'elements': elements,
                        }
                    )
                pbar.close()

            return attr_elements

    def __fetch_attribute_elements_chunks(self, future_session, limit):
        """Fetch add'l rows from this object instance from the Intelligence
        Server"""
        return [
            cubes.cube_single_attribute_elements_coroutine(
                future_session,
                cube_id=self._id,
                attribute_id=attribute['id'],
                offset=0,
                limit=limit,
            )
            for attribute in self.attributes
        ]

    def unpublish(self, force: bool = False) -> bool:
        """Unpublish Cube by removing all of its caches.

        Args:
            force: If True, then no additional prompt will be shown before
                deleting Cube.

        Returns:
            True when all of cube's caches were deleted and it was successfully
            unpublished. False otherwise.
        """
        user_input = 'N'
        if not force:
            user_input = (
                input(
                    f"Are you sure you want to unpublish cube '{self.name}' with ID: "
                    f"{self._id}? This operation will delete all of its caches. "
                    f"[Y/N]: "
                )
                or 'N'
            )
        if force or user_input == 'Y':
            all_caches_num = len(self.caches)
            deleted_caches_num = 0
            for cache in self.caches:
                if cache.delete(True):
                    deleted_caches_num += 1
            if all_caches_num == deleted_caches_num:
                logger.info(
                    f"Cube '{self.name}' with ID: '{self._id}' was successfully "
                    f"unpublished."
                )
                self._caches = []
                return True
            else:
                logger.warning(
                    f"Cube '{self.name}' with ID: '{self._id}' was not successfully "
                    f"unpublished, because not all of its caches were deleted."
                )
                return False
        else:
            return False

    def _prepare_instance_body(self) -> dict:
        return self._instance_config._request_body()

    @property
    def attributes(self) -> list[dict]:
        if not self.__definition_retrieved:
            self._get_definition()
        return self._attributes

    @property
    def metrics(self) -> list[dict]:
        if not self.__definition_retrieved:
            self._get_definition()
        return self._metrics

    @property
    def attr_elements(self) -> list:
        if not self.__definition_retrieved:
            self._get_definition()
        if not self._attr_elements and self._id:
            if self._parallel is True:
                # TODO: move the fallback inside the function to apply
                # per-attribute, like with non-async version.
                self._attr_elements = fallback_on_timeout()(
                    self.__get_attr_elements_async
                )(50000)[0]
            else:
                self._attr_elements = self.__get_attr_elements()
            self._instance_config._populate_attr_elements(self._attr_elements)
        return self._attr_elements

    @property
    def _instance_config(self):
        if not self.__definition_retrieved:
            self._get_definition()
        if self.__filter is None:
            self.__filter = Filter(
                attributes=self._attributes,
                metrics=self._metrics,
                attr_elements=self._attr_elements,
            )
        return self.__filter

    @property
    def selected_attributes(self):
        """Selected attributes for filtering."""
        return self._instance_config.attr_selected

    @property
    def selected_metrics(self):
        """Selected metrics for filtering."""
        return self._instance_config.metr_selected

    @property
    def selected_attr_elements(self):
        """Selected attribute elements for filtering."""
        return self._instance_config.attr_elem_selected

    @property
    def dataframe(self) -> DataFrame:
        if self._dataframe is None:
            exception_handler(
                msg="Dataframe not loaded. Retrieve with Report.to_dataframe().",
                exception_type=Warning,
            )
        return self._dataframe

    @property
    def dataframes(self):
        if len(self._dataframes) == 0:
            exception_handler(
                msg="Dataframe not loaded. Retrieve with Report.to_dataframe().",
                exception_type=Warning,
            )
        return self._dataframes

    @property
    def table_definition(self):
        return self._table_definition

    @property
    def caches(self):
        if not self._caches:
            self.get_caches()
        return self._caches



# This example is for a dashboard user to modify data in a Rally workspace using
# a modern grid configured with transaction functions. The workflow will rely on this
# script to make actual modifications in the defect. The properties available for
# modification in the defect are `Name`, `Description`, `Priority`, `Owner`,
# `Iteration`, `TargetDate`, `State` and `PlanEstimate`. Only the `Name` property is
# required. Property  `FormattedID` cannot be modified.

# To connect with Rally API we use the library 'pyral'. Python library to handle Rally
# connection is not installed on the environment by default so it is necessary to choose
# runtime which have such library installed in the settings of the script. Additionally
# it is necessary to specify correct 'Network Access' to allow script to communicate
# with some other IP then IP of the environment
from pyral import Rally, restapi
from datetime import datetime

# Helper method to return datetime object in correct format for Rally operations
def parse_date_time_value(dateValue):
    # somehow it will be converted to this datetime format
    #   "%Y-%m-%d %H:%M:%S.%f" -> 2023-04-01 00:00:00.000
    date_formats = [
        "%Y-%m-%d %H:%M:%S.%f",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%d",
        "%m/%d/%Y %H:%M:%S.%f",
        "%m/%d/%Y %H:%M:%S",
        "%m/%d/%Y",
    ]
    parsed_datetime = None
    for date_format in date_formats:
        try:
            parsed_datetime = datetime.strptime(dateValue, date_format)
            break
        except Exception:
            pass

    if parsed_datetime is None:
        raise Exception("Wrong Type of the datatime input value convertion!!!!")
    else:
        print(parsed_datetime)
        return parsed_datetime

try:
    # create connection to Rally. If you want to use other way then providing
    # api key please check documentation of this method 
    rally = Rally(
        server=$rally_server,
        apikey=$api_key,
        workspace=$workspace,
        project=$project,
    )

    # deal with the every row data via prepared statements
    # we assume that the logic of passing data from dashboard works correctly and each
    # transactional column is either the array of length which is equal to number
    # of rows passed from dashboard or `None`.
    for index in range(len($formatted_id)):
        # DELETE RALLY DEFECT
        tmp_formatted_id = $formatted_id[index]
        # first it is necessary to find object which will be deleted
        defect = rally.get(
            "Defect",
            fetch=True,
            query=f"FormattedID = {tmp_formatted_id}",
            projectScopeDown=True,
        ).next()

        rally.delete('Defect', defect.ObjectID)

except restapi.RallyRESTAPIError as error:
    raise Exception(f"Error has occurred: {error}")
# this error is raised from `rally.get` when the object was not found and we are trying
# to iterate with `next` method
except StopIteration as error:
    raise Exception(f"Error while iterating has ocurred: {error}")



from enum import auto

from mstrio.modeling.schema.helpers import SchemaObjectReference
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable


class RelationshipType(AutoName):
    """Enumeration constants used to specify type of the relationship"""

    ONE_TO_MANY = auto()
    ONE_TO_ONE = auto()
    MANY_TO_MANY = auto()


class Relationship(Dictable):
    """Python representation of MicroStrategy Attributes Relationship object.

    Attributes:
        parent: parent in a relationship
        child: child in a relationship
        joint_child: list of joint children in a relationship
        relationship_table: relationship table in a relationship
        ralationship_type: type of relationship, one of:
            one_to_many, one_to_one, many_to_many
    """

    # For backwards compatibility, to avoid deprecation.
    RelationshipType = RelationshipType

    _FROM_DICT_MAP = {
        'parent': SchemaObjectReference.from_dict,
        'child': SchemaObjectReference.from_dict,
        'joint_child': [SchemaObjectReference.from_dict],
        'relationship_table': SchemaObjectReference.from_dict,
        'relationship_type': RelationshipType,
    }

    def __init__(
        self,
        relationship_type: RelationshipType,
        relationship_table: SchemaObjectReference,
        parent: SchemaObjectReference,
        child: SchemaObjectReference | None = None,
        joint_child: list[SchemaObjectReference] | None = None,
    ) -> None:
        self.relationship_type = relationship_type
        self.relationship_table = relationship_table
        self.parent = parent
        if (child and joint_child) or (child is None and joint_child is None):
            raise AttributeError(
                "Please specify either 'child' or 'joint_child' parameter in the "
                "constructor."
            )
        elif child:
            self.child = child
        elif joint_child:
            self.joint_child = (
                joint_child if isinstance(joint_child, list) else [joint_child]
            )

    def __eq__(self, other):
        if not isinstance(other, Relationship):
            # don't attempt to compare against unrelated types
            return False

        if hasattr(self, 'child') and hasattr(other, 'child'):
            return (
                self.relationship_type == other.relationship_type
                and self.relationship_table == other.relationship_table
                and self.parent == other.parent
                and self.child == other.child
            )
        elif hasattr(self, 'joint_child') and hasattr(other, 'joint_child'):
            return (
                self.relationship_type == other.relationship_type
                and self.relationship_table == other.relationship_table
                and self.parent == other.parent
                and bool(set(self.joint_child).intersection(other.joint_child))
            )
        else:
            return False

    def __repr__(self):
        if hasattr(self, 'child'):
            child_name = self.child.name
        else:
            child_name = "[" + ", ".join(child.name for child in self.joint_child) + "]"
        return (
            f'{self.parent.name} -> {self.relationship_type.name} -> {child_name} '
            f'relationship'
        )



from mstrio.connection import get_connection
from mstrio.object_management.folder import Folder
from mstrio.helpers import AggregatedRights
from mstrio.users_and_groups.user import User
from mstrio.users_and_groups.user_group import list_user_groups

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# assign ACL AggregatedRights.VIEW to UserGroup
user_group = list_user_groups(conn, name='Mobile')[0]
user_group.acl_add(
    rights=AggregatedRights.VIEW,
    trustees=[User(conn, name='Administrator')],
    denied=False,
)

# assign ACL AggregatedRights.VIEW to UserGroup and propagate it to children
user_group = list_user_groups(conn, name='Mobile')[0]
user_group.acl_add(
    rights=AggregatedRights.VIEW,
    trustees=[User(conn, name='Administrator')],
    denied=False,
    propagate_to_children=True,
)

# assign ACLs for folder
folder = Folder(conn, id='945A968E40E4EA2C9114BBBE885DBC24')
folder.acl_add(rights=127, trustees=[User(conn, name='Administrator')], denied=False)

# assign ACLs for folder and propagate to children
folder = Folder(conn, id='945A968E40E4EA2C9114BBBE885DBC24')
folder.acl_add(
    rights=127,
    trustees=[User(conn, name='Administrator')],
    denied=False,
    propagate_to_children=True,
)



from typing import TYPE_CHECKING

from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@ErrorHandler(err_msg="Error getting members of security filter with ID {id}")
def get_security_filter_members(
    connection: 'Connection', id: str, project_id: str = None, error_msg: str = None
):
    """Get the users and user groups that the specified security filter is
    applied to."""
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id
    return connection.get(
        endpoint=f'/api/securityFilters/{id}/members',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error updating members of security filter with ID {id}")
def update_security_filter_members(
    connection: 'Connection',
    id: str,
    body: dict,
    project_id: str = None,
    error_msg: str = None,
    throw_error: bool = True,
):
    """Update members information for a specific security filter."""
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id
    return connection.patch(
        endpoint=f'/api/securityFilters/{id}/members',
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@unpack_information
@ErrorHandler(err_msg="Error creating new security filter.")
def create_security_filter(
    connection: 'Connection',
    body: dict,
    show_filter_tokens: bool = False,
    show_expression_as: str = None,
    error_msg: str = None,
    throw_error: bool = True,
):
    """Creates a new security filter in the changeset,
    based on the definition provided in request body.

    Args:
        connection: MicroStrategy REST API connection object
        body (dict): Security Filter creation body
        error_msg (str, optional): Custom Error Message for Error Handling
        show_expression_as (str, optional): specify how expressions should be
            presented
            Available values:
                - `tree` (default)
                - `tokens`
        show_filter_tokens (bool, optional): specify whether `qualification`
            is returned in `tokens` format, along with `text` and `tree`
            format

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            endpoint='/api/model/securityFilters',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=body,
            params={
                'showFilterTokens': str(show_filter_tokens).lower(),
                'showExpressionAs': show_expression_as,
            },
        )


@unpack_information
@ErrorHandler(err_msg="Error getting security filter {id} definition.")
def get_security_filter(
    connection: 'Connection',
    id: str,
    project_id: str = None,
    changeset_id: str = None,
    show_expression_as: str = None,
    show_fields: str = None,
    show_filter_tokens: bool = False,
    error_msg: str = None,
):
    """Get the definition of a security filter.
    The project ID is required to return a security filter's definition
    in metadata. The changeset ID is required to return a security filter's
    definition within a specific changeset. To execute the request,
    either the project ID or changeset ID needs to be provided.
    If both are provided, only the changeset ID is used.

    Args:
        connection: MicroStrategy REST API connection object
        id (str): Security Filter ID. The ID can be:
            - the object ID used in the metadata.
            - the object ID used in the changeset, but not yet committed
            to metadata.
        project_id (str, optional): Project ID
        changeset_id (str, optional): Changeset ID
        show_expression_as (str, optional): specify how expressions should be
            presented
            Available values:
                - `tree` (default)
                - `tokens`
        show_fields: Specifies what additional information to return.
            Only 'acl' is supported.
        show_filter_tokens (bool, optional): specify whether `qualification`
            is returned in `tokens` format, along with `text` and `tree`
            format
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        endpoint=f'/api/model/securityFilters/{id}',
        headers={'X-MSTR-ProjectID': project_id, 'X-MSTR-MS-Changeset': changeset_id},
        params={
            'showExpressionAs': show_expression_as,
            'showFields': show_fields,
            'showFilterTokens': str(show_filter_tokens).lower(),
        },
    )


@unpack_information
@ErrorHandler(err_msg="Error updating security filter with ID {id}")
def update_security_filter(
    connection: 'Connection',
    id: str,
    body: dict,
    show_expression_as: str = None,
    show_fields: str = None,
    show_filter_tokens: bool = False,
    error_msg: str = None,
    throw_error: bool = True,
):
    """Updates a specific security filter in the changeset,
    based on the definition provided in the request body.

    Args:
        connection: MicroStrategy REST API connection object
        id (str): Security Filter ID. The ID can be:
            - the object ID used in the metadata.
            - the object ID used in the changeset, but not yet committed
            to metadata.
        body (dict): Security Filter update info
        show_expression_as (str, optional): specify how expressions should be
            presented
            Available values:
                - `tree` (default)
                - `tokens`
        show_fields: Specifies what additional information to return.
            Only 'acl' is supported.
        show_filter_tokens (bool, optional): specify whether `qualification`
            is returned in `tokens` format, along with `text` and `tree`
            format
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.put(
            endpoint=f'/api/model/securityFilters/{id}',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={
                'showExpressionAs': show_expression_as,
                'showFields': show_fields,
                'showFilterTokens': str(show_filter_tokens).lower(),
            },
            json=body,
        )


@ErrorHandler(err_msg="Error getting information for set of security filters.")
def get_security_filters(
    connection: 'Connection',
    project_id: str = None,
    name_contains: str = None,
    offset: int = 0,
    limit: int = -1,
    fields: str = None,
    error_msg: str = None,
):
    """Get all list of Security Filters for a project.
    You can set the offset and limit for pagination function.

    Args:
        connection: MicroStrategy REST API connection object
        project_id (string, optional): id of project
        name_contains (string, optional): text that security filter's name
            must contain
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
        fields: top-level field whitelist which allows client to selectively
            retrieve part of the response model
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint='/api/securityFilters',
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'nameContains': name_contains,
            'offset': offset,
            'limit': limit,
            'fields': fields,
        },
    )



import logging
from enum import Enum, auto
from typing import TYPE_CHECKING, Union

from packaging import version

from mstrio import config
from mstrio.api import monitors
from mstrio.connection import Connection
from mstrio.helpers import MstrException, PartialSuccess, Success, VersionException
from mstrio.server import Node, Project
from mstrio.utils.entity import Entity, EntityBase
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import validate_param_value
from mstrio.utils.monitors import all_nodes_async
from mstrio.utils.time_helper import DatetimeFormats, map_str_to_datetime
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.users_and_groups import User

logger = logging.getLogger(__name__)

ISERVER_VERSION_11_3_2 = '11.3.0200'


class DeliveryType(AutoName):
    RESERVED = auto()
    EMAIL = auto()
    FILE = auto()
    PRINTER = auto()
    CUSTOM = auto()
    INBOX = auto()
    CLIENT = auto()
    CACHE = auto()
    MOBILE = auto()
    BLACKBERRY = auto()
    IPHONE = auto()
    IPAD = auto()
    CAMPAIGN = auto()
    SNAPSHOT = auto()
    FTP = auto()
    ANDROID = auto()
    LAST_ONE = auto()
    COUNT = auto()
    ALL_INCLUDING_SNAPSHOT = auto()


class SortByV1(Enum):
    JOB_TYPE = 'jobType'
    USER_FULL_NAME = 'userFullName'
    OBJECT_ID = 'objectId'
    STATUS = 'status'
    PROJECT_ID = 'projectId'
    # check if valid as there are multiple
    # inconsistent information about sorting
    PROJECT_NAME = 'projectName'
    CREATION_TIME = 'creationTime'
    DESCRIPTION = 'description'
    JOB_ID = 'jobId'


class JobStatus(AutoName):
    READY = auto()
    EXECUTING = auto()
    WAITING = auto()
    COMPLETED = auto()
    ERROR = auto()
    CANCELING = auto()
    STOPPED = auto()
    WAITING_ON_GOVERNOR = auto()
    WAITING_FOR_AUTOPROMPT = auto()
    WAITING_FOR_PROJECT = auto()
    WAITING_FOR_CACHE = auto()
    WAITING_FOR_CHILDREN = auto()
    WAITING_FOR_RESULTS = auto()
    LOADING_PROMPT = auto()
    RESOLVING_DESTINATION = auto()
    DELIVERING = auto()
    EXPORTING = auto()
    CACHE_READY = auto()
    WAITING_FOR_DI_FILE = auto()
    WAITING_FOR_CONFLICT_RESOLVE = auto()
    STEP_PAUSING = auto()


class JobType(AutoName):
    INTERACTIVE = auto()
    SUBSCRIPTION = auto()
    PREDICTIVE_CACHE = auto()


# NOTE: new endpoint Enums below:
class ObjectType(AutoName):
    OTHERS = auto()
    REPORT = auto()
    CUBE = auto()
    DOCUMENT = auto()
    DOSSIER = auto()
    DASHBOARD = 'dossier'


class PUName(AutoName):
    BROWSING = auto()
    RESOLUTION = auto()
    QUERY_EXECUTIONS = 'query_execution'
    ANALYTICAL = auto()
    SQL_ENGINE = auto()
    DATA_FORMATTING = auto()
    NCS = auto()
    REST_ASYNCHRONOUS = auto()


class SubscriptionType(AutoName):
    RESERVED = auto()
    EMAIL = auto()
    PRINTER = auto()
    CUSTOM = auto()
    INBOX = auto()
    CLIENT = auto()
    CACHE = auto()
    MOBILE = auto()
    MOBILE_BLACKBERRY = auto()
    MOBILE_IPHONE = auto()
    MOBILE_IPAD = auto()
    CAMPAIGN = auto()
    SNAPSHOT = auto()
    FTP = auto()
    MOBILE_ANDROID = auto()
    LAST_ONE = auto()
    COUNT = auto()
    ALL = auto()
    ALL_INCLUDING_SNAPSHOT = auto()


class SortBy(Enum):
    ID_ASC = '+id'
    ID_DESC = '-id'
    TYPE_ASC = '+type'
    TYPE_DESC = '-type'
    STATUS_ASC = '+status'
    STATUS_DESC = '-status'
    USER_ASC = '+user'
    USER_DESC = '-user'
    DESCRIPTION_ASC = '+description'
    DESCRIPTION_DESC = '-description'
    OBJECT_TYPE_ASC = '+objectType'
    OBJECT_TYPE_DESC = '-objectType'
    OBJECT_ID_ASC = '+objectId'
    OBJECT_ID_DESC = '-objectId'
    SUBSCRIPTION_TYPE_ASC = '+subscriptionType'
    SUBSCRIPTION_TYPE_DESC = '-subscriptionType'
    PROCESSING_UNIT_PRIORITY_ASC = '+processingUnitPriority'
    PROCESSING_UNIT_PRIORITY_DESC = '-processingUnitPriority'
    CREATION_TIME_ASC = '+creationTime'
    CREATION_TIME_DESC = '-creationTime'
    COMPLETED_TASKS_ASC = '+completedTasks'
    COMPLETED_TASKS_DESC = '-completedTasks'
    PROJECT_ID_ASC = '+projectId'
    PROJECT_ID_DESC = '-projectId'
    PROJECT_NAME_ASC = '+projectName'
    PROJECT_NAME_DESC = '-projectName'
    SUBSCRIPTION_RECIPIENT_ASC = '+subscriptionRecipient'
    SUBSCRIPTION_RECIPIENT_DESC = '-subscriptionRecipient'
    MEMORY_USAGE_ASC = '+memoryUsage'
    MEMORY_USAGE_DESC = '-memoryUsage'
    ELAPSED_TIME_ASC = '+elapsedTime'
    ELAPSED_TIME_DESC = '-elapsedTime'


@method_version_handler('11.3.0200')
def _set_api_wrappers(connection) -> dict:
    if version.parse(connection.iserver_version) == version.parse(
        ISERVER_VERSION_11_3_2
    ):
        return {
            (
                'id',
                'description',
                'status',
                'type',
                'priority',
                'processing_unit_priority',
                'warehouse_priority',
                'object_id',
                'user',
                'project_name',
                'creation_time',
                'total_tasks',
                'completed_tasks',
                'filter_name',
                'template_name',
                'sql',
                'subscription_owner',
                'subscription_recipient',
                'destination',
            ): monitors.get_job
        }
    else:
        return {
            (
                'id',
                'type',
                'status',
                'user',
                'description',
                'object_type',
                'object_id',
                'parent_id',
                'child_ids',
                'subscription_type',
                'processing_unit_priority',
                'step_id',
                'pu_name',
                'creation_time',
                'completed_tasks',
                'total_tasks',
                'project_name',
                'subscription_recipient',
                'memory_usage',
                'elapsed_time',
                'step_elapsed_time',
                'filter_name',
                'template_name',
                'sql',
                'subscription_owner',
                'error_time',
                'error_message',
                'step_statistics',
            ): monitors.get_job_v2
        }


@method_version_handler('11.3.0200')
def list_jobs(
    connection: "Connection",
    node: Node | str | None = None,
    user: Union["User", str] | None = None,
    description: str | None = None,
    type: JobType | str | None = None,
    status: JobStatus | str | None = None,
    object_id: str | None = None,
    object_type: ObjectType | None = None,
    project: Project | str | None = None,
    pu_name: PUName | str | None = None,
    subscription_type: SubscriptionType | str | None = None,
    subscription_recipient: Union["User", str] | None = None,
    memory_usage: str | None = None,
    elapsed_time: str | None = None,
    sort_by: SortBy | str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["Job"] | list[dict]:
    """List jobs objects or job dictionaires.
    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`
        node(Node, str, optional): Node object or name, if not passed list jobs
            on all nodes
        user(User, str, optional): Field to filter on job owner's full name
        (exact match)
        description(str, optional): Field to filter on job description (partial
            match)
        type(JobType, optional): Field to filter on job type (exact match),
        status(JobStatus, optional): Field to filter on job status (exact match)
        object_id(str, optional): Field to filter on object type (exact
            match)
        object_type(ObjectType, optional): Field to filter on object type (exact
            match)
        project(Project, str, optional): Field to filter on project name (exact
            match)
        pu_name(PUName, optional): Field to filter on processing unit name
            (exact match)
        subscription_type(SubscriptionType, optional): Field to filter on
        subscription type (exact match)
        subscription_recipient(User, str, optional): Field to filter on
            subscription recipient's full name (exact match)
        memory_usage(str, optional): Field to filter on the job elapsed time,
            for example 'gt:100' means filtering jobs with memory usage greater
            than to 100 MB. Valid operators are:
            gt - greater than
            lt - less than
        elapsed_time(str, optional): Field to filter on the job elapsed time,
            for example 'gt:100' means filtering jobs with elapsed time greater
            than 100 milliseconds(11.3.3+) or seconds(11.3.2). Valid operators:
            gt - greater than
            lt - less than
        sort_by(SortBy, optional): Specify sorting criteria, for example
            SortBy.STATUS_ASC means sorting status is ascending order
            or SortBy.USER_DESC means sorting user in descending order.
            Currently, the server supports sorting only by single field.
        to_dictionary(bool, optional): if True, return Schedules as
            list of dicts
        limit(int, optional): maximum number of schedules returned
        **filters(optional): Local filter parameters

    Examples:
        >>> list_jobs_v1(connection, duration='gt:100')

    Returns:
        List[Job] | List[dict]: list of Job objects or dictionaries.
    """
    user = user.full_name if isinstance(user, Entity) else user
    subscription_recipient = (
        subscription_recipient.full_name
        if isinstance(user, Entity)
        else subscription_recipient
    )
    if project:
        project = (
            project
            if isinstance(project, Project)
            else Project(connection, name=project)
        )

    # depending on version call either one or the other
    if version.parse(connection.iserver_version) == version.parse(
        ISERVER_VERSION_11_3_2
    ):
        filters = __prepare_v1_request(
            description,
            object_type,
            pu_name,
            subscription_type,
            subscription_recipient,
            memory_usage,
            elapsed_time,
            filters,
        )
        return list_jobs_v1(
            connection=connection,
            node=node,
            project=project,
            status=status,
            job_type=type,
            user=user,
            limit=limit,
            to_dictionary=to_dictionary,
            **filters,
        )
    else:
        project_name = project.name if project else None
        node_name = node.name if isinstance(node, Node) else node
        msg = "Error fetching chunk of jobs."
        objects = all_nodes_async(
            connection,
            async_api=monitors.get_jobs_v2_async,
            filters=filters,
            error_msg=msg,
            unpack_value='jobs',
            limit=limit,
            node_name=node_name,
            user=user,
            description=description,
            type=type,
            object_id=object_id,
            object_type=object_type,
            project_name=project_name,
            status=status,
            pu_name=pu_name,
            subscription_type=subscription_type,
            subscription_recipient=subscription_recipient,
            memory_usage=memory_usage,
            elapsed_time=elapsed_time,
            sort_by=sort_by,
            fields=['jobs'],
        )

    if to_dictionary:
        return objects
    else:
        return [Job.from_dict(source=obj, connection=connection) for obj in objects]


@method_version_handler('11.3.0200')
def list_jobs_v1(
    connection: Connection,
    node: Node | str | None = None,
    project: Project | str | None = None,
    status: JobStatus | str | None = None,
    job_type: JobType | str | None = None,
    user: Union["User", str] | None = None,
    object_id: str | None = None,
    sort_by: SortBy | str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["Job"] | list[dict]:
    """List job objects or job dictionaries. Optionally filter list.
    NOTE: list_jobs can return up to 1024 jobs per request.

    Args:
        connection(object): MicroStrategy connection object returned by
            'connection.Connection()'
        node(Node, str, optional): Node object or name, if not passed list jobs
            on all nodes
        project(str, Project, optional): Project id or object to filter by
        status(JobStatus, optional): Job status to filter by
        job_type(JobType, optional): Job type to filter by
        user(User, str, optional): User object or full name to filter by
        object_id(str, optional): Object id to filter by
        sort_by(SortBy, optional): Specifies sorting criteria to sort by
        to_dictionary(bool, optional): if True, return Schedules as
            list of dicts
        limit(int, optional): maximum number of schedules returned.
        **filters: Available filter parameters:[id, description, status,
            jobType, duration, jobId, objectd]
            - duration(str, optional): Field to filter on the job elapsed time,
                for example 'gt:100' means filtering jobs with elapsed time
                greater than to 100 seconds. Valid operators:
                gt - greater than
                lt - less than

    Examples:
        >>> list_jobs_v1(connection, duration='gt:100')

    Returns:
        List[Job] | List[dict]: list of Job objects or dictionaries.
    """
    project_id = project.id if isinstance(project, Project) else project
    user_full_name = user.full_name if isinstance(user, Entity) else user

    if 'duration' in filters:
        filters['duration'] = __elapsed_filtering(filters['duration'])

    node_name = node.name if isinstance(node, Node) else node
    msg = "Error fetching chunk of jobs."
    objects = all_nodes_async(
        connection,
        async_api=monitors.get_jobs_async,
        filters=filters,
        error_msg=msg,
        unpack_value='jobs',
        limit=limit,
        node_name=node_name,
        project_id=project_id,
        status=status,
        job_type=job_type,
        user_full_name=user_full_name,
        object_id=object_id,
        sort_by=sort_by,
    )

    if to_dictionary:
        return objects
    else:
        return [Job.from_dict(source=obj, connection=connection) for obj in objects]


@method_version_handler('11.3.0200')
def kill_jobs(
    connection: Connection, jobs: list[Union["Job", str]]
) -> Success | PartialSuccess | MstrException:
    """Kill existing jobs by Job objects or job ids.

    Args:
        connection(object): MicroStrategy connection object returned by
            'connection.Connection()'
        jobs: List of Job objects or job ids to kill

    Returns:
        Success: object if all jobs were killed
        PartialSuccess: if not all jobs were killed
        MstrException: otherwise
    """
    validate_param_value("jobs", jobs, list)
    jobs = [job.id if isinstance(job, Job) else job for job in jobs]
    return monitors.cancel_jobs(connection, jobs)


@method_version_handler('11.3.0200')
def kill_all_jobs(
    connection: Connection,
    user: Union["User", str] | None = None,
    description: str | None = None,
    type: JobType | str | None = None,
    status: JobStatus | str | None = None,
    object_id: str | None = None,
    object_type: ObjectType | str | None = None,
    project: Project | str | None = None,
    pu_name: PUName | str | None = None,
    subscription_type: SubscriptionType | str | None = None,
    subscription_recipient: Union["User", str] | None = None,
    memory_usage: str | None = None,
    elapsed_time: str | None = None,
    force: bool = False,
    **filters,
) -> Success | PartialSuccess | MstrException:
    """Kill jobs filtered by passed fields

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`
         node(Node, str, optional): Node object or name, if not passed kill jobs
            on all nodes
        user(User, str, optional): Field to filter on job owner's full name
        (exact match)
        description(str, optional): Field to filter on job description (partial
            match)
        type(JobType, optional): Field to filter on job type (exact match),
        status(JobStatus, optional): Field to filter on job status (exact match)
        object_id(str, optional): Field to filter on object id (exact
            match)
        object_type(ObjectType, optional): Field to filter on object type (exact
            match)
        project(Project, str, optional): Field to filter on project name (exact
            match)
        pu_name(PUName, optional): Field to filter on processing unit name
            (exact match)
        subscription_type(SubscriptionType, optional): Field to filter on
        subscription type (exact match)
        subscription_recipient(User, str, optional): Field to filter on
            subscription recipient's full name (exact match)
        memory_usage(str, optional): Field to filter on the job elapsed time,
            for example 'gt:100' means filtering jobs with memory usage greater
            than to 100 MB. Valid operators are:
            gt - greater than
            lt - less than
        elapsed_time(str, optional): Field to filter on the job elapsed time,
            for example 'gt:100' means filtering jobs with elapsed time greater
            than 100 milliseconds(11.3.3+) or seconds(11.3.2). Valid operators:
            gt - greater than
            lt - less than
        force(bool, optional): Force flag

    Returns:
        Success: object if all jobs were killed
        PartialSuccess: if not all jobs were killed
        MstrException: otherwise
    """

    jobs = list_jobs(
        connection,
        user,
        description,
        type,
        status,
        object_id,
        object_type,
        project,
        pu_name,
        subscription_type,
        subscription_recipient,
        memory_usage,
        elapsed_time,
        filters,
    )
    assert jobs, "No jobs to kill"

    jobs_ids = [job.id for job in jobs]
    if not force:
        print(f"Found jobs: {jobs}")
        user_input = input("Are you sure you want to kill those jobs?[Y/N]: ") or 'N'
    if force or user_input == 'Y':
        return kill_jobs(connection, jobs_ids)


def __prepare_v1_request(
    description: str,
    object_type: str,
    pu_name: str,
    subscription_type: str,
    subscription_recipient: str,
    memory_usage: str,
    elapsed_time: str,
    filters: dict,
) -> dict:
    # raise VersionException if parameter not supported in v1
    unsupported_parameters = [
        description,
        object_type,
        pu_name,
        subscription_type,
        subscription_recipient,
        memory_usage,
    ]
    params_str = (
        "description, object_type, pu_name, subscription_type, "
        "subscription_recipient, memory_usage"
    )
    if any(unsupported_parameters):
        msg = (
            f"Passed unsupported parameter for this version of IServer "
            f"({ISERVER_VERSION_11_3_2}) Parameters supported only in 11.3.3+ "
            f"versions: {params_str}"
        )
        raise VersionException(msg)

    if elapsed_time:
        filters['duration'] = elapsed_time
    return filters


def __elapsed_filtering(elapsed: str) -> str:
    if (
        isinstance(elapsed, str)
        and len(elapsed) > 3
        and (elapsed.startswith('gt:') or elapsed.startswith('lt:'))
    ):
        if elapsed.startswith('gt'):
            return f'>{elapsed[3:]}'
        elif elapsed.startswith('lt'):
            return f'<{elapsed[3:]}'
        else:
            return elapsed
    else:
        raise TypeError('Incorrect "duration" format, correct example: "gt:100" ')


@class_version_handler('11.3.0200')
class Job(EntityBase):
    """Python representation of a Job object.
    Note: Some functionality is missing when working with 11.3.2 I-Server
        versions. To use all functionality 11.3.3 and up will be required.
        There are also some differences between 11.3.2 and 11.3.3+ Job
        attributes values.

    Attributes:
        connection: A MicroStrategy connection object
        id: Job information id
        description: Description of the job
        status: Status of the job
        type: Type of the job
        object_id: Id of object
        object_type: Type of object
        parent_id: Parent ID
        childs_ids: Array of children IDs
        subscription_recipient: Subscription recipient of the job
        subscription_type: Subscription delivery type of the job
        processing_unit_priority: Job Processing Unit Priority
        step_id: Step ID of the job
        user: Full name of the job initiator
        project_name: Name of the project in which the job is active
        project_id: Id of the project in which the job is active
        pu_name: Processing Unit name
        creation_time: Creation time of the job
        elapsed_time: Time elapsed since the creation of the job in milliseconds
            (11.3.3+) or seconds (11.3.2)
        step_elapsed_time: Time elapsed since the creation of the step
            in milliseconds
        memory_usage: Memory usage of the job in bytes
        client_machine: Client machine running the job
        total_tasks: Total tasks of the job
        completed_tasks: Completed tasks of the job
        error_message: Error message for the job
        error_time: Error time of the job
        filter_name: Name of the filter for the job
        sql: SQL of the job
        subscription_owner: Subscription owner
        template_name: Name of the template for the job
        step_statistics: Step statistics for the job
    """

    _FROM_DICT_MAP = {
        'creation_time': DatetimeFormats.FULLDATETIME,
        'error_time': DatetimeFormats.FULLDATETIME,
        'type': JobType,
        'status': JobStatus,
        'object_type': ObjectType,
        'subscription_type': SubscriptionType,
        'pu_name': PUName,
    }
    _API_GETTERS = {
        (
            'id',
            'type',
            'status',
            'user',
            'description',
            'object_type',
            'object_id',
            'parent_id',
            'child_ids',
            'subscription_type',
            'processing_unit_priority',
            'step_id',
            'pu_name',
            'creation_time',
            'completed_tasks',
            'total_tasks',
            'project_name',
            'subscription_recipient',
            'memory_usage',
            'elapsed_time',
            'step_elapsed_time',
            'filter_name',
            'template_name',
            'sql',
            'subscription_owner',
            'error_time',
            'error_message',
            'step_statistics',
        ): monitors.get_job_v2
    }
    _REST_ATTR_MAP = {
        "job_id": None,  # delete job_id and only use id
        "job_type": "type",
        "user_full_name": "user",
        "total_task": "total_tasks",
        "completed_task": "completed_tasks",
        "filter": "filter_name",
        "template": "template_name",
        "machine": "client_machine",
        "priority": None,
        "warehouse_priority": None,
        "destination": None,
        "project_id": None,
        "username": None,
        "delivery_type": None,
        "duration": "elapsed_time",
    }

    def __init__(self, connection: Connection, id: str) -> None:
        """Initialize the Job object, populates it with I-Server data.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            id: Job information id
        """
        Job._API_GETTERS = _set_api_wrappers(connection)
        super().__init__(connection, object_id=id)

    def _init_variables(self, **kwargs):
        """Initialize variables given kwargs.

        Note: attributes not accepted by any implementation of this function
            in the inheritance chain will be disregarded.
        """
        kwargs = self._rest_to_python(kwargs)
        # create _AVAILABLE_ATTRIBUTES map
        self._AVAILABLE_ATTRIBUTES.update(
            {key: type(val) for key, val in kwargs.items()}
        )

        self._connection = kwargs.get('connection')
        self._id = kwargs.get('id')
        self._description = kwargs.get('description')
        self._status = JobStatus(kwargs.get('status')) if kwargs.get('status') else None
        self._processing_unit_priority = kwargs.get('processing_unit_priority')
        self._creation_time = map_str_to_datetime(
            "creation_time", kwargs.get("creation_time"), self._FROM_DICT_MAP
        )
        self._elapsed_time = kwargs.get('elapsed_time')
        self._project_name = kwargs.get('project_name')
        self._object_id = kwargs.get('object_id')
        self._object_type = (
            ObjectType(kwargs.get('object_type')) if kwargs.get('object_type') else None
        )
        self._sql = kwargs.get('sql')
        self._subscription_owner = kwargs.get('subscription_owner')
        self._subscription_recipient = kwargs.get('subscription_recipient')
        self._client_machine = kwargs.get('client_machine')
        self._type = JobType(kwargs.get('type')) if kwargs.get('type') else None
        self._user = kwargs.get('user')
        self._total_tasks = kwargs.get('total_tasks')
        self._completed_tasks = kwargs.get('completed_tasks')
        self._parent_id = kwargs.get('parent_id')
        self._child_ids = kwargs.get('child_ids')
        self._subscription_type = (
            SubscriptionType(kwargs.get('subscription_type'))
            if kwargs.get('subscription_type')
            else kwargs.get('subscription_type')
        )
        self._step_id = kwargs.get('step_id')
        self._pu_name = PUName(kwargs.get('pu_name')) if kwargs.get('pu_name') else None
        self._memory_usage = kwargs.get('memory_usage')
        self._step_elapsed_time = kwargs.get('step_elapsed_time')
        self._filter_name = kwargs.get('filter_name')
        self._template_name = kwargs.get('template_name')
        self._error_time = (
            map_str_to_datetime(
                "error_time", kwargs.get("error_time"), self._FROM_DICT_MAP
            )
            if kwargs.get("error_time")
            else None
        )
        self._error_message = kwargs.get('error_message')
        self._step_statistics = kwargs.get('step_statistics')

    def kill(self) -> bool:
        """Kill the job.

        Returns:
            True if successfully killed job, False otherwise.
        """
        response = monitors.cancel_job(self._connection, str(self.id))
        if config.verbose:
            success_msg = f"Successfully killed {self.id}."
            fail_msg = f"Error killing job {self.id}."
            msg = success_msg if response.ok else fail_msg
            logger.info(msg)
        return response.ok

    def list_properties(self) -> dict:
        """List Job object properties."""
        properties = super().list_properties()
        properties.pop('connection', None)
        return properties

    @property
    def status(self):
        return self._status

    @property
    def description(self):
        return self._description

    @property
    def processing_unit_priority(self):
        return self._processing_unit_priority

    @property
    def creation_time(self):
        return self._creation_time

    @property
    def project_name(self):
        return self._project_name

    @property
    def object_type(self):
        return self._object_type

    @property
    def object_id(self):
        return self._object_id

    @property
    def sql(self):
        return self._sql

    @property
    def subscription_owner(self):
        return self._subscription_owner

    @property
    def subscription_recipient(self):
        return self._subscription_recipient

    @property
    def client_machine(self):
        return self._client_machine

    @property
    def user(self):
        return self._user

    @property
    def total_tasks(self):
        return self._total_tasks

    @property
    def completed_tasks(self):
        return self._completed_tasks

    @property
    def parent_id(self):
        return self._parent_id

    @property
    def child_ids(self):
        return self._child_ids

    @property
    def subscription_type(self):
        return self._subscription_type

    @property
    def step_id(self):
        return self._step_id

    @property
    def pu_name(self):
        return self._pu_name

    @property
    def memory_usage(self):
        return self._memory_usage

    @property
    def elapsed_time(self):
        return self._elapsed_time

    @property
    def step_elapsed_time(self):
        return self._step_elapsed_time

    @property
    def filter_name(self):
        return self._filter_name

    @property
    def template_name(self):
        return self._template_name

    @property
    def error_time(self):
        return self._error_time

    @property
    def error_message(self):
        return self._error_message

    @property
    def step_statistics(self):
        return self._step_statistics



import logging
from collections.abc import Callable
from dataclasses import dataclass
from enum import auto

from pandas import DataFrame

from mstrio import config
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable
from mstrio.utils.version_helper import class_version_handler

logger = logging.getLogger(__name__)


class VldbPropertyType(AutoName):
    """Enumeration constant indicating type of VLDB property value."""

    INT32 = auto()
    INT64 = auto()
    DOUBLE = auto()
    BOOLEAN = auto()
    STRING = auto()
    DATE = auto()
    TIME = auto()


@dataclass
class PreviewOption(Dictable):
    """Class representation of single SQL Preview Option related to VldbSetting.

    Attributes:
        value (str): Value of SQL preview option.
        sql_preview (str): Description of SQL preview option
    """

    value: str
    sql_preview: str


class ResolvedLocation(AutoName):
    """Enumeration constant indicating type of VLDB property resolved
    and next resolved locations."""

    REPORT_TARGET = auto()
    REPORT = auto()
    TEMPLATE_TARGET = auto()
    TEMPLATE = auto()
    OBJECT = auto()
    PROJECT = auto()
    DB_ROLE = auto()
    DBMS = auto()
    DEFAULT = auto()


@dataclass
class VldbSetting(Dictable):
    """Class representation of single MicroStrategy VLDB setting.

    Attributes:
        type (VldbPropertyType): Type of VLDB setting value.
        name (str): VLDB setting name.
        property_set (str): Name of corresponding VLDB setting set.
        group_id (int): ID of VLDB setting group, settings are sorted by that
            parameter while listing.
        group_name (str): Name of VLDB setting group.
        value (str | int | float | bool): Current value set for VLDB setting.
        default_value (str | int | float | bool): Default value.
        display_type (str): Determines way of displaying settings on frontend.
        display_name (str): Name that will be displayed on frontend for
            particular settings.
        sql_preview (bool): Determines if SQL preview should be displayed.
        options (list[PreviewOption]): List of allowed sql options.
        resolved_location (ResolvedLocation): Type of location where value of
            related VLDB setting was last time resolved.
        is_inherited (bool): Whether the related VLDB setting value is inherited
            from other objects or not. If False value is set on the object.
            If True, value is inherited from other objects.
        max_value (int, optional): Shows maximum number of options that can be
            provided for the property, if applicable.
        next_value (str, optional): Previously set value.
        next_resolved_location (ResolvedLocation, optional): Previously set
            resolved_location.
        value_separator (str, optional): Values will be joined using that symbol
            when displaying on frontend.
    """

    _FROM_DICT_MAP = {
        'type': VldbPropertyType,
        'options': [PreviewOption.from_dict],
        'resolved_location': ResolvedLocation,
        'next_resolved_location': ResolvedLocation,
    }

    type: VldbPropertyType
    property_set: str
    name: str
    group_id: int
    group_name: str
    value: str | int | float | bool
    default_value: str | int | float | bool
    display_type: str
    display_name: str
    sql_preview: bool
    options: list[PreviewOption]
    resolved_location: ResolvedLocation
    is_inherited: bool
    max_value: int | None = None
    next_value: str | None = None
    next_resolved_location: ResolvedLocation | None = None
    value_separator: str | None = None

    type_map = {
        VldbPropertyType.INT32: int,
        VldbPropertyType.INT64: int,
        VldbPropertyType.DOUBLE: float,
        VldbPropertyType.BOOLEAN: lambda x: x == 'true',
        VldbPropertyType.STRING: str,
        VldbPropertyType.DATE: str,
        VldbPropertyType.TIME: str,
    }

    def __repr__(self) -> str:
        value_repr, default_value_repr = str(self.value), str(self.default_value)
        if isinstance(self.value, str):
            value_repr, default_value_repr = (
                f"'{value_repr}'",
                f"'{default_value_repr}'",
            )
        setting_repr = (
            f"VldbSetting(type={self.type}, property_set='{self.property_set}', "
            f"name='{self.name}', group_id={self.group_id}, "
            f"group_name='{self.group_name}', value={value_repr}, "
            f"default_value={default_value_repr}, "
        )
        if self.max_value:
            setting_repr += f"max_value={self.max_value}, "
        setting_repr += (
            f"display_type='{self.display_type}', "
            f"display_name='{self.display_name}', "
            f"sql_preview={self.sql_preview}, options={self.options}, "
            f"resolved_location={self.resolved_location}, "
            f"is_inherited={self.is_inherited})"
        )
        return setting_repr

    @classmethod
    def from_dict(cls, source: dict) -> 'VldbSetting':
        source['value'] = cls.type_map[VldbPropertyType(source['type'])](
            source['value']
        )
        return super().from_dict(source=source)

    def list_properties(self) -> dict:
        """List properties for the VLDB setting."""
        return self.to_dict()


@class_version_handler('11.3.0800')
class ModelVldbMixin:
    """ModelVldbMixin class adds VLDB management for supported objects through
    modeling service endpoints.

    Class attributes:
        _MODEL_VLDB_API (dict[str, Callable]): Dict containing references to API
            wrappers.

    Attributes:
        vldb_settings (dict[str, VldbSetting]): Dict with settings names as keys
            and VldbSetting objects as values.

    Objects currently supporting ModelVldbMixin settings are Project,
    DatasourceInstance and OlapCube. Must be mixedin with Entity or its
    subclasses.
    """

    _MODEL_VLDB_API: dict[str, Callable]
    _vldb_settings: dict[str, VldbSetting] = {}

    def _fetch(self):
        applicable_properties = self._MODEL_VLDB_API['GET_APPLICABLE'](
            self.connection, self.id
        ).json()['applicableProperties']
        advanced_properties = self._MODEL_VLDB_API['GET_ADVANCED'](
            self.connection, self.id
        ).json()['advancedProperties']['vldbProperties']
        self._vldb_settings = {}

        for key, applicable in applicable_properties.items():
            vldb_setting = VldbSetting.from_dict(
                {
                    'propertySet': ''.join(key.split('.')[0][1:-1]),
                    'sqlPreview': applicable['showSqlPreview'],
                    **advanced_properties[key],
                    **applicable,
                }
            )
            self._vldb_settings[vldb_setting.name] = vldb_setting

    def __filter_settings(
        self,
        vldb_settings: dict[str, VldbSetting],
        str_arg: str,
        args: list[str],
    ) -> dict[str, VldbSetting]:
        all_args = {
            getattr(setting, str_arg) for setting in self.vldb_settings.values()
        }

        for arg in args:
            if arg not in all_args:
                msg = f"There is no VldbSetting with '{str_arg}' equals to '{arg}'."
                raise ValueError(msg)

        return {
            name: setting
            for name, setting in vldb_settings.items()
            if getattr(setting, str_arg) in args
        }

    def list_vldb_settings(
        self,
        set_names: list[str] | None = None,
        names: list[str] | None = None,
        groups: list[int] | list[str] | None = None,
        to_dictionary: bool = False,
        to_dataframe: bool = False,
    ) -> dict[str, VldbSetting]:
        """List VLDB settings according to given parameters.

        Args:
            set_names (list[str], optional): List of names of VLDB settings
                sets.
            names (list[str], optional): List of names of VLDB settings.
            groups (list[int], list[str], optional): List of group names or ids.
            to_dictionary(bool, optional): If True, return VldbSetting
                objects as list of dicts, default False.
            to_dataframe (bool, optional): If True, return VldbSetting
                objects as pandas DataFrame, default False.

        Raises:
            ValueError: If there are no VldbSetting objects with given
                parameters or some of the parameters are incorrectly
                specified.
            TypeError: If groups list parameter consists of objects of
                different types.

        Returns:
            Dict with settings names as keys and VldbSetting objects as values.
        """

        if to_dictionary and to_dataframe:
            msg = (
                "Please select either `to_dictionary=True` or `to_dataframe=True`, "
                "but not both.",
            )
            raise ValueError(msg)

        self._fetch()
        vldb_settings = self.vldb_settings.copy()

        if set_names:
            vldb_settings = self.__filter_settings(
                vldb_settings, 'property_set', set_names
            )

        if groups:
            if all(isinstance(group, int) for group in groups):
                vldb_settings = self.__filter_settings(
                    vldb_settings, 'group_id', groups
                )
            elif all(isinstance(group, str) for group in groups):
                vldb_settings = self.__filter_settings(
                    vldb_settings, 'group_name', groups
                )
            else:
                msg = "Elements in the groups list must be of the same type."
                raise TypeError(msg)

        if names:
            vldb_settings = self.__filter_settings(vldb_settings, 'name', names)

        if not vldb_settings:
            msg = "There is no VldbSetting objects found with given parameters."
            raise ValueError(msg)

        if to_dictionary or to_dataframe:
            vldb_settings = {
                key: setting.to_dict() for key, setting in vldb_settings.items()
            }

        if to_dataframe:
            return DataFrame(vldb_settings)
        else:
            return vldb_settings

    def alter_vldb_settings(
        self, names_to_values: dict[str, str | int | float | bool]
    ) -> None:
        """Alter VLDB settings according to given name to value pairs.

        Note:
            Only common value type conversion will be done by default,
            such as int->float, bool->int, int->bool and only if that is
            possible according to interval of allowed values, error will be
            thrown in other cases.

        Args:
            names_to_values (dict[str, str | int | float | bool]): Dict with
                VlDB settings names as keys and newly requested to set values
                as dictionary values.

        Raises:
            ValueError: If there are no VldbSetting objects for some
                provided names or names and values are not provided or some
                of provided values will not be possible to set.
            TypeError: If some of provided values to set have wrong type,
                different from type of default VLDB setting value.

        Returns:
            None
        """

        if not names_to_values:
            msg = "Please provide not empty dict with names to values pairs."
            raise ValueError(msg)

        body = {'advancedProperties': {'vldbProperties': {}}}

        for name, value in names_to_values.items():
            if name not in self.vldb_settings:
                msg = f"There is no VldbSetting with name '{name}'."
                raise ValueError(msg)

            setting = self.vldb_settings[name]

            if isinstance(value, int):
                value = self.__convert_int(value, setting)

            if not isinstance(value, type(setting.value)):
                value_type = type(setting.value)
                msg = f"Wrong type of VldbSetting value, it should be {value_type}."
                raise TypeError(msg)

            key = f'[{setting.property_set}].[{setting.name}]'
            body['advancedProperties']['vldbProperties'][key] = {'value': value}

        self._MODEL_VLDB_API['PUT_ADVANCED'](self.connection, self.id, body)

        self._fetch()
        if config.verbose:
            logger.info(f"VLDB settings for object with ID '{self.id}' were altered.")

    def __convert_int(self, value, setting):
        setting_value_type = type(setting.value)
        max_value = setting.max_value if setting.max_value else float('inf')
        min_value = 0 if setting.max_value else float('-inf')

        if (value < min_value or value > max_value) or (
            not setting.max_value and isinstance(value, bool)
        ):
            msg = "Incorrect value of VldbSetting, it should be"
            if setting.max_value:
                msg += f" positive and less or equal to {setting.max_value}"
            msg += f" of {setting_value_type}."
            raise ValueError(msg)

        return setting_value_type(value)

    def reset_vldb_settings(
        self,
        set_names: list[str] | None = None,
        names: list[str] | None = None,
        groups: list[int] | list[str] | None = None,
        force: bool = False,
    ) -> None:
        """Reset VLDB settings specified by parameters to its default values.
        If call was without parameters then additional prompt will be shown.

        Args:
            set_names (list[str], optional): List of names of VLDB setting
                sets.
            names (list[str], optional): List of names of VLDB settings.
            groups (list[int], list[str], optional): List of groups names
                or ids.
            force: (bool, optional): If True, no additional prompt will
                be shown before resetting in case when all other
                arguments are not provided, default False.

        Raises:
            ValueError: If there are no VldbSetting objects with given
                parameters or some of the parameters are incorrectly
                specified.
            TypeError: If groups list parameter consists of objects of
                different types.

        Returns:
            None
        """

        if not (set_names or names or groups or force):
            user_input = input(
                f"Are you sure you want to reset all VLDB settings for object with "
                f"ID '{self.id}'? [Y/N]: "
            )
            if user_input != 'Y':
                return

        body = {'advancedProperties': {'vldbProperties': {}}}
        names = self.list_vldb_settings(set_names, names, groups)

        for name in names:
            setting = self.vldb_settings[name]
            if setting.value != setting.default_value:
                key = f'[{setting.property_set}].[{setting.name}]'
                body['advancedProperties']['vldbProperties'][key] = {'value': None}

        self._MODEL_VLDB_API['PUT_ADVANCED'](self.connection, self.id, body)

        self._fetch()
        if config.verbose:
            logger.info(f"VLDB settings for object with ID '{self.id}' were reset.")

    @property
    def vldb_settings(self) -> dict[str, VldbSetting]:
        if not self._vldb_settings:
            self._fetch()
        return self._vldb_settings



from mstrio.api import tables as tables_api
from mstrio.connection import Connection


def update_structure(
    connection: Connection,
    id: str,
    column_merge_option: str | None = None,
    ignore_table_prefix: bool | None = None,
):
    data = tables_api.update_structure(
        connection, id, column_merge_option, ignore_table_prefix
    ).json()

    data.update(data.pop('information', {}))
    data.update({'id': data.pop('objectId', None)})

    return data
